# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Defines some utility routines used to map a provided set of quantities.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings

from IPython import embed

import numpy
from scipy import ndimage

from astropy.wcs import WCS

from matplotlib import pyplot, patches
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import MaxNLocator



##############################################################################

def masked_pixelized_image(x, y, z, pixelscale=1.0, zmin=None, zmax=None, imshow_prep=False,
                           fill_value=0.0):
    r"""

    Provided a set of pixelized data, return a masked image with a type
    `numpy.ma.MaskedArray`_.  The :math:`x` coordinates are organized
    along rows and the :math:`y` coordinates are organized along
    columns.  I.e., i.e., img[0,1] = :math:`z(x_0,y_1)`, unless
    *imshow_prep* is requested (see below).
    
    Args:
        x (numpy.ndarray): X coordinates of the pixels
        y (numpy.ndarray): Y coordinates of the pixels
        z (numpy.ndarray): Image values at :math:`x,y`.
        pixelscale (float): (Optional) Pixelscale of the image in
            arcsec/pixel.
        zmin (float): (Optional) Minimum z value to include in the
            output image.  Default is to allow all pixel values.
        zmax (float): (Optional) Maximum z value to include in the
            output image.  Default is to allow all pixel values.
        imshow_prep (bool): (Optional) Prepare the matrix for use with
            `pyplot.imshow`_.  If *imshow_prep* is True, before output,
            the matrix is reordered such that increasing :math:`x`
            values are along columns and increasing :math:`y` values are
            along rows; i.e., the output is the transpose of the default
            behavior.  The appropriate call to imshow that will then put
            increasing x values along the abcissa and increasing y
            values along the ordinate is, e.g.::

                ext, img = masked_pixelized_image(x, y, z, imshow_prep=True)
                pyplot.imshow(img, interpolation='nearest', extent=ext, origin='lower')

            Note that origin **must** be "lower" when calling
            `pyplot.imshow`_ on the array produced by this
            routine for the :math:`x,y` ordering to be as expected!
        fill_value (float): (Optional) The default value to use for
            pixels without any data.

    Returns:
        numpy.ndarray, `numpy.ma.MaskedArray`_: The first object
        returned is a four-element array with the extent of the image
        from the bottom edge to the top edge of the x and y pixels,
        respectively.  The second object returned is the image data and
        associated (boolean) mask.
    """

    # Image extent
    ext = numpy.empty(4, dtype=float)
    ext[0], ext[1] = numpy.amin(x)-pixelscale/2., numpy.amax(x)+pixelscale/2.
    ext[2], ext[3] = numpy.amin(y)-pixelscale/2., numpy.amax(y)+pixelscale/2.

    # Image size
    nx, ny = numpy.floor((ext[1]-ext[0])/pixelscale), numpy.floor((ext[3]-ext[2])/pixelscale)
#    print('MAP:  nx,ny: {0},{1}'.format(nx, ny))
#    print('MAP: extent: {0}'.format(ext))

    # Image and mask values
    img = numpy.full((nx,ny), fill_value, dtype=float)
    i = numpy.floor((x.ravel()-ext[0])/pixelscale).astype(int)
    j = numpy.floor((y.ravel()-ext[2])/pixelscale).astype(int)
    i[i == nx] = nx-1
    j[j == ny] = ny-1
#    print(i)
#    print(j)
    img[i,j] = z
    mask = numpy.full((nx,ny), True, dtype=bool)
    mask[i,j] = False

    # Mask values
    if zmin is not None:
        mask[(img < zmin)] = True
    if zmax is not None:
        mask[(img > zmax)] = True

#    # Flip the x axis (sky-right coordinates?)
#    if xflip:
#        img = numpy.fliplr(img)
#        mask = numpy.fliplr(mask)
#        ext[0], ext[1] = ext[1], ext[0]
#        print('EXTENT: {0}'.format(ext))

    # Prep the image for direct display using pyplot.imshow
    if imshow_prep:
        img = numpy.transpose(img)
        mask = numpy.transpose(mask)

    return ext, numpy.ma.MaskedArray(img, mask=mask, fill_value=fill_value)


def map_quantity(x, y, z, zmin=None, zmax=None, ncolors=64, dots=False, cmap='coolwarm',
                 colorbar=False, nticks=7, clabel=None, flux=None, fixpdf=False, xlim=None,
                 xlabel=None, ylim=None, ylabel=None, pixelscale=None, fill_value=0.0,
                 contour_levels=None, **kwargs):

    """
    Copyright (C) 2013-2014, Michele Cappellari
    E-mail: cappellari_at_astro.ox.ac.uk
    http://purl.org/cappellari/software
    
    *Revision history*:
        | V1.0: Michele Cappellari, Paranal, 11 November 2013
        | V1.0.1: Clip values before contouring. MC, Oxford, 26 February
            2014
        | V1.0.2: Include SAURON colormap. MC, Oxford, 29 January 2014
        | V1.0.3: Call set_aspect(1). MC, Oxford, 22 February 2014
        | V1.0.4: Call autoscale_view(tight=True). Overplot small dots
            by default.  MC, Oxford, 25 February 2014
        | V1.0.5: Use axis('image'). MC, Oxford, 29 March 2014
        | V1.0.6: Allow changing colormap. MC, Oxford, 29 July 2014
        | V1.0.7: Added optional fixpdf keyword to remove PDF artifacts
            like `stackoverflow_15822159`_ .  Make nice tick levels for
            colorbar. Added nticks keyword for colorbar.  MC, Oxford, 16
            October 2014

    .. _stackoverflow_15822159: http://stackoverflow.com/questions/15822159/aliasing-when-saving-matplotlib-filled-contour-plot-to-pdf-or-eps
 
    """
    if zmin is None:
        zmin = numpy.min(z)
    if zmax is None:
        zmax = numpy.max(z)

    if cmap is None:
        cmap = 'coolwarm'

#    print('raveling')

    x, y, z = map(numpy.ravel, [x, y, z])
#    print('getting axes')
    ax = pyplot.gca()

    # Provide a pixelized map
    if pixelscale is not None:
#        print('pixelizing')
        ext, img = masked_pixelized_image(x, y, z, pixelscale, zmin=zmin, zmax=zmax,
                                          fill_value=fill_value, imshow_prep=True)

#        print('made pixelized image')
        cs = ax.imshow(img, interpolation='nearest', cmap=cmap, vmin=zmin,
                       vmax=zmax, extent=ext, origin='lower')
#        print('output imshow')

        if contour_levels is not None:
            nx = img.shape[0]
            dx = (ext[1]-ext[0])/nx
            ny = img.shape[1]
            dy = (ext[3]-ext[2])/ny
            ximg, yimg = numpy.meshgrid(ext[0]+dx*(numpy.arange(nx)+0.5),
                                        ext[2]+dy*(numpy.arange(ny)+0.5), indexing='xy')
            cs1 = ax.contour(ximg, yimg, img, levels=contour_levels, colors='k', linewidths=1.5)
       
    # Provide an interpolated map
    else:
        levels = numpy.linspace(zmin, zmax, ncolors) 
        cs = ax.tricontourf(x, y, z.clip(zmin, zmax), levels=levels, cmap=cmap)

    ax.axis('image')  # Equal axes and no rescaling
    ax.minorticks_on()
    ax.tick_params(length=10, which='major')
    ax.tick_params(length=5, which='minor')
#    print('axes limits: {0}'.format(ax.get_xlim()))
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if flux is not None:
        ax.tricontour(x, y, -2.5*numpy.log10(flux/numpy.max(flux).ravel()),
                      levels=numpy.arange(20), colors='k') # 1 mag contours

    if fixpdf:  # remove white contour lines in PDF at expense of larger file size
        ax.tricontour(x, y, z.clip(zmin, zmax), levels=levels, zorder=0,
                      cmap=cmap)

    if dots:
        ax.plot(x, y, '.k', markersize=kwargs.get("markersize", 3))

    if colorbar:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        ticks = MaxNLocator(nticks).tick_values(zmin, zmax)
        cbar = pyplot.colorbar(cs, cax=cax, ticks=ticks)
        if clabel:
            cbar.set_label(clabel)

#    print('done map')

    return cs

# TODO: Move this into datacube?
def map_extent(hdu, ext, offset=True):
    """
    Get the on-sky extent of a map using the provided WCS coordinates.

    Args:
        hdu (`astropy.io.fits.HDUList`_): Fits HDU list.
        ext (str): Extension of the file with the map.
        offset (bool): (**Optional**)  Return the extent as the offset
            from the coordinates of the object in arcseconds.  The WCS
            is assumed to provide RA and declination in units of
            degrees, and the header must contain the coordinates of the
            object with keywords `OBJRA` and `OBJDEC`.  If offset is
            True and these keywords do not exist, the function will
            throw a warning and proceed *without* applying any offset.
            Default is just to return the extent of the WCS coordinates.

    Returns:
        list: List of four floats: minimum and maximum x and minimum and
        maximum y.

    """
    if offset:
        try:
            objra = hdu['PRIMARY'].header['OBJRA']
            objdec = hdu['PRIMARY'].header['OBJDEC']
        except Exception as e:
            print(e)
            warnings.warn('OBJRA and/or OBJDEC in the header of the PRIMARY extension.  '
                          'Offseting to the center pixel.')
            objra, objdec = None, None

    # Does the extension contain a cube or a single map image
    cube = len(hdu[ext].data.shape) > 2

    # Get the shape of the image
    if cube:
        nc,ny,nx = hdu[ext].data.shape
    else:
        ny,nx = hdu[ext].data.shape

    # Create a grid with the pixel coordinates
    x,y = numpy.meshgrid( numpy.arange(nx)+1, numpy.arange(ny)+1,
                          indexing='ij')

    # Reshape them into an array with X,Y coordinates
    xy = numpy.array([x.reshape(nx*ny), y.reshape(nx*ny)]).T
    if cube:
        xy = numpy.append(xy, numpy.ones(nx*ny).reshape(-1,1), axis=1)

    # Read the World Coordinate System information from the fits header
    wcs = WCS(header=hdu[ext].header, fix=False)
    if wcs.naxis == 3 and not cube:
        warnings.warn('WCS is for 3D, but provided data is 2D.  Attempting to drop the last axis.')
        wcs = wcs.dropaxis(-1)

    # Convert the pixel coordinates into world coordinates
    XY = wcs.all_pix2world(xy, 1)
    x = XY[:,0].reshape(nx,ny)
    y = XY[:,1].reshape(nx,ny)

    # So now the x coordinate of:
    #  hdu[ext].data[:,j,i]
    # is
    #  x[j,i]
    if not offset:
        return [ numpy.amax(x), numpy.amin(x), numpy.amin(y), numpy.amax(y) ]

    if objra is None or objdec is None:
        center = tuple([_n // 2 for _n in x.shape])
        objra = x[center]
        objdec = y[center]

    # Convert the world coordinates to offset from center in arcsec
    x = (x-objra)*numpy.cos(numpy.radians(objdec))*3600.
    y = (y-objdec)*3600.

    # Return the extent
    return [numpy.amax(x), numpy.amin(x), numpy.amin(y), numpy.amax(y)]


def map_center_pixel_offset(hdu_1, hdu_2, ext, quiet=False):
    """
    Determine the offset in pixels between the object center in two the
    maps of two fits files.

    Args:
        hdu_1 (`astropy.io.fits.HDUList`_): Fits HDU list
            for the reference map.
        hdu_2 (`astropy.io.fits.HDUList`_): Fits HDU list
            for the comparison map.
        ext (str): Extension *in both files* with the maps.
        quiet (bool): (**Optional**)  Suppress terminal output.
    
    Returns:
        float: Two floats with the offset in x and y between the two maps.
    """
    cube = len(hdu_1[ext].data.shape) > 2

    objra = hdu_1['PRIMARY'].header['OBJRA']
    objdec = hdu_1['PRIMARY'].header['OBJDEC']

    if not quiet:
        print('Coordinates of 1st object: {0:10.6f} {1:9.5f}'.format(objra, objdec))
        print('Coordinates of 2nd object: {0:10.6f} {1:9.5f}'.format(
                                                                hdu_2['PRIMARY'].header['OBJRA'],
                                                                hdu_2['PRIMARY'].header['OBJDEC']))

    wcs_1 = WCS(header=hdu_1[ext].header, fix=False)
    wcs_2 = WCS(header=hdu_2[ext].header, fix=False)

    XY = numpy.zeros(3 if cube else 2).reshape(1,-1)
    XY[0,0] = objra
    XY[0,1] = objdec

    xy_1 = wcs_1.all_world2pix(XY, 1)
    xy_2 = wcs_2.all_world2pix(XY, 1)
    if not quiet:
        print('Object pixel coordinates in 1st HDU: {0:6.2f} {1:6.2f}'.format(xy_1[0,0], xy_1[0,1]))
        print('Object pixel coordinates in 2nd HDU: {0:6.2f} {1:6.2f}'.format(xy_2[0,0], xy_2[0,1]))
        print('Pixel shifts: {0} {1}'.format(xy_2[0,0]-xy_1[0,0], xy_2[0,1]-xy_1[0,1]))

    return (xy_2[0,0]-xy_1[0,0]), (xy_2[0,1]-xy_1[0,1])


def match_map_arrays(arr1, ext1, arr2, ext2, dx, dy, tol=1e-6, truncate=True, pixelscale=0.5):
    """
    Match the centers and extent of two map arrays by applying the
    previously calculated pixel offsets.  See
    :func:`map_center_pixel_offset`.

    This is just a wrapper for the two supporting functions below.

    Both maps *must* be two dimensional; i.e., a single map, not a set
    of maps arranged in different channels.

    Args:
        arr1 (`numpy.ma.MaskedArray`_): Reference map data array.
        ext1 (list): Reference map extent in arcseconds.
        arr2 (`numpy.ma.MaskedArray`_): Comparison map data array.
        ext1 (list): Comparison map extent in arcseconds.
        dx (float): **Pixel** offset in the *first* dimension (x).
        dy (float): **Pixel** offset in the *second* dimension (y).
        tol (float): (**Optional**) Tolerance for non-integer pixel
            shifts.
        truncate (float): (**Optional**) Truncate the shape of the
            shifted array to match the unshifted array.  If False, the
            unshifted array is padded to match the shifted array.
        pixelscale (float): (**Optional**) The pixel scale (extent units
            per array pixel) that *must be* common to both input arrays.

    Returns:
        numpy.ma.MaskedArray, list: The two matched arrays and the
        extent that is common to both.

    """
    # Must apply a sub-pixel shift
    if abs(dx-numpy.around(dx))>tol or abs(dy-numpy.around(dy))>tol:
        return _match_map_arrays_sub_pixel_shift(arr1, ext1, arr2, ext2, -dx, -dy,
                                                 truncate=truncate, pixelscale=pixelscale) \
                            if arr1.size > arr2.size else \
                    _match_map_arrays_sub_pixel_shift(arr2, ext2, arr1, ext1, dx, dy, swap=True,
                                                      truncate=truncate, pixelscale=pixelscale)
    # Can just apply a whole pixel shift
    return _match_map_arrays_integer_pixel_shift(arr1, ext1, arr2, ext2, -int(numpy.around(dx)),
                                                 -int(numpy.around(dy))) \
                            if arr1.size > arr2.size else \
                    _match_map_arrays_integer_pixel_shift(arr2, ext2, arr1, ext1,
                                                          int(numpy.around(dx)),
                                                          int(numpy.around(dy)), swap=True)


def _match_map_arrays_sub_pixel_shift(arr1, ext1, arr2, ext2, dx, dy, swap=False, truncate=False,
                                      pixelscale=0.5):

    # Pad the second array with an even number of pixels
    add = int(abs(dx))+1 if abs(dx) > abs(dy) else int(abs(dy))+1
    add *= 2
    if arr2.shape[0]+add - arr1.shape[0] > 0 and (arr2.shape[0]+add - arr1.shape[0]) % 2 > 1:
        add += 2
    newshape = tuple(numpy.array(arr2.shape) + add)

    # Pad and copy the array and its mask
    _arr2 = numpy.ma.masked_all(newshape, dtype=float)
    _arr2[add//2:add//2+arr2.shape[0],add//2:add//2+arr2.shape[1]] = arr2[:,:]
    _gpm = numpy.invert(numpy.ma.getmaskarray(_arr2)).astype(float)
    # Shifts are in pixels, so adjust given the new coordinates
    dx -= add//2
    dy -= add//2
    newext2 = list(numpy.array(ext2) + numpy.array([add/2, -add/2, -add/2, add/2])*pixelscale)

    # Get the input and output coordinates
    newx, newy = numpy.meshgrid(numpy.arange(newshape[0]), numpy.arange(newshape[1]), indexing='ij')
    outcoo = numpy.array([newx.ravel(), newy.ravel()])
    incoo = numpy.array([newx.ravel()-dx, newy.ravel()-dy])
    indx = ~(incoo[0,:] < 0) & ~(incoo[1,:] < 0)

    # Shift the second array and set the mask
    shifted_arr2 = numpy.ma.masked_all(newshape, dtype=float)
    shifted_arr2[outcoo[0,indx], outcoo[1,indx]] = \
        ndimage.map_coordinates(_arr2.filled(fill_value=0.0), incoo[:,indx], order=1)
    shifted_gpm = numpy.zeros(newshape, dtype=float)
    shifted_gpm[outcoo[0,indx], outcoo[1,indx]] = \
        ndimage.map_coordinates(_gpm, incoo[:,indx], order=1)
    shifted_arr2[ shifted_gpm < 0.8 ] = numpy.ma.masked

    # Return the two arrays
    sub = arr1.shape[0]-shifted_arr2.shape[0]
    if sub == 0:
        return (shifted_arr2, arr1, ext1) if swap else (arr1, shifted_arr2, ext1)
    # Pad the shifted array up to the shape of the first array
    if sub > 0:
        _shifted_arr2 = numpy.ma.masked_all(arr1.shape, dtype=float)
        _shifted_arr2[:shifted_arr2.shape[0],:shifted_arr2.shape[1]] = shifted_arr2[:,:]
        return (_shifted_arr2, arr1, ext1) if swap else (arr1, _shifted_arr2, ext1)

    sub *= -1

    # If truncation is requested, set the shifted 2nd array to the same
    # size as the first input array
    if truncate:
        _shifted_arr2 = numpy.ma.masked_all(arr1.shape, dtype=float)
        _shifted_arr2[:,:] = shifted_arr2[:arr1.shape[0],:arr1.shape[1]]
        return (_shifted_arr2, arr1, ext1) if swap else (arr1, _shifted_arr2, ext1)

    # Pad the first array to the size of the shifted array
    _arr1 = numpy.ma.masked_all(shifted_arr2.shape, dtype=float)
    _arr1[:arr1.shape[0],:arr1.shape[1]] = arr1[:,:]
    return (shifted_arr2, _arr1, newext2) if swap else (_arr1, shifted_arr2, newext2)


# Arr1 must be bigger or the same size as Arr2
def _match_map_arrays_integer_pixel_shift(arr1, ext1, arr2, ext2, dx, dy, swap=False):
    if arr1.size < arr2.size:
        raise ValueError('First array must be larger than the second.')
    if abs(dx) > (arr1.shape[0] - arr2.shape[0])/2 or abs(dy) > (arr1.shape[1] - arr2.shape[1])/2:
        raise ValueError('First array is not large enough to accommodate second array and shift.')
    _arr = numpy.ma.MaskedArray(numpy.zeros(arr1.shape, dtype=float),
                                mask=numpy.ones(arr1.shape, dtype=bool))
    _arr[dx:dx+arr2.shape[0],dy:dy+arr2.shape[1]] = arr2[:,:]
    return (_arr, arr1, ext1) if swap else (arr1, _arr, ext1) 


def map_beam_patch(extent, ax, fwhm=2.5, pos=(0.1,0.1), **kwargs):
    width = extent[0]-extent[1]
    return patches.Circle(pos, fwhm/width/2, transform=ax.transAxes, **kwargs)



def permute_wcs_axes(wcs, axes):
    r"""
    Permute the axes in an `astropy.wcs.WCS`_ object.

    The number of axes must match the number axes defined by the
    `astropy.wcs.WCS`_ object. Axes are iteratively swapped until
    they're in the requested order. For example, to transpose the
    axes of a 3D WCS, set ``axes=[2,1,0]``.

    .. note::

        Method uses the :func:`deepcopy` method of the
        `astropy.wcs.WCS`_ object when altering and returning a new
        object. If the axes are already sorted, this function returns
        a deepcopy of the input ``wcs``.

    Args:
        wcs (`astropy.wcs.WCS`_):
            Object with the world-coordinate system.
        axes (array-like):
            New locations for the current axes; i.e., the axis
            currently at index ``i`` (0-indexed) is moved to be at
            index ``axis[i]``. For example, to swap the axes for a 2D
            WCS, set ``axes=[1,0]``.

    Returns:
        `astropy.wcs.WCS`_: The world-coordinate system with
        re-ordered axes.

    Raises:
        ValueError:
            Raised if the length of the ``axes`` vector is not the
            same as the number of axes defined by the
            `astropy.wcs.WCS`_ object, or if ``axes`` includes
            undefined axes (i.e., a value less than 0 or >= the
            number of axes in the WCS.)
    """
    if len(axes) != wcs.wcs.naxis:
        raise ValueError('Number of specified axes must match the WCS object.')
    if not numpy.array_equal(numpy.arange(len(axes)), numpy.sort(axes)):
        raise ValueError('Must select defined axes; 0 ... wcs.wcs.naxis-1.')
    _wcs = wcs.deepcopy()
    if numpy.array_equal(axes, numpy.sort(axes)):
        return _wcs
    _axes = numpy.atleast_1d(axes).copy()
    for i,a in enumerate(_axes):
        if a == i:
            continue
        _wcs = _wcs.swapaxes(a,i)
        _axes[i], _axes[a] = i, a
    return _wcs




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Implements a full-spectrum fitting algorithm that fits a
wavelength-dependent dispersion.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import time
import warnings

from IPython import embed

import numpy
from scipy import optimize

import astropy.constants

from ppxf.ppxf_util import gaussian_filter1d
from ppxf.capfit import cov_err

from ..util.filter import interpolate_masked_vector
from ..util.sampling import angstroms_per_pixel, Resample
from ..util.bitmask import BitMask
from ..util import modeling


class PieceWiseLinear:
    def __init__(self, deg, c=None, rng=[-1,1]):
        self.deg = deg
        self.np = self.deg
        self.coeff = numpy.ones(self.np, dtype=float) if c is None else c
        self.rng = rng

    def reset_range(self, rng):
        self.rng = rng

    def __call__(self, x, a=None):
        if a is not None:
            if len(a) != self.np:
                raise ValueError('Incorrect number of parameters')
            self.coeff = a
        if self.deg == 1:
            return self.coeff[0]
        return interpolate.interp1d(numpy.linspace(*self.rng, self.deg), self.coeff,
                                    fill_value='extrapolate', assume_sorted=True)(x)


class LegendrePolynomial:
    def __init__(self, deg, c=None, rng=None):
        self.deg = deg
        self.np = self.deg+1
        self.coeff = numpy.ones(self.np, dtype=float) if c is None else c
        self.rng = rng

    def reset_range(self, rng):
        self.rng = rng

    def __call__(self, x, a=None):
        if a is not None:
            if len(a) != self.np:
                raise ValueError('Incorrect number of parameters')
            self.coeff = a
        return numpy.polynomial.legendre.legval(x, self.coeff)


# TODO:  assess how rng should be used here
class ScaledEmpirical:
    def __init__(self, x, y, scale_func=None, rng=None, assume_sorted=True):
        self.deg = 0 if scale_func is None else scale_func.deg
        self.np = 1 if scale_func is None else scale_func.np
        self.sf = LegendrePolynomial(0) if scale_func is None else scale_func
        self.coeff = numpy.array([1.0]) if scale_func is None else scale_func.coeff
        self.reset_range(rng)
        self.interp = interpolate.interp1d(modeling.scaled_coordinates(x, rng=self.rng), y,
                                           fill_value='extrapolate', assume_sorted=assume_sorted)

    def reset_rng(self, rng):
        if hasattr(self.sf, 'reset_range'):
            self.sf.reset_range(rng)
        self.rng = rng

    def __call__(self, x, a=None):
        if a is not None:
            if len(a) != self.np:
                raise ValueError('Incorrect number of parameters')
            self.coeff = a
        return self.sf(x, a=self.coeff)*self.interp(x)
        

class RukiaBitMask(BitMask):
    """
    Global mask bits used for Rukia.
    """
    def __init__(self):
        super(RukiaBitMask, self).__init__(['LOSIGMA', 'HISIGMA'])


class Rukia:
    """
    Fit a spectrum by convolving a *single* template spectrum with a
    Gaussian kernel that has a wavelength-dependent width.

    This instantiation defines up to three functional forms used to
    define the fitted model. The objects used to describe each
    functional form must have an ``np`` attribute that provides the
    number of parameters and it must be callable with a form::

        func(x, a=a)

    where ``func`` is the object, ``x`` is the coordinate vector, and
    ``a`` are the model parameters. For example, see
    :class:`LegendrePolynomial`.

    The total number of parameters for the fitted model is the sum of
    the number of parameters for each functional form, plus one
    additional parameter if a Doppler shift is also included in the
    fit (see ``fit_shift``).

    Args:
        sigma_model (object):
            An object that provides the parametrized form of sigma as
            a function of wavelength.
        add_model (object, optional):
            An object that provides the parametrized form of an
            additive function included in the model.
        mul_model (object, optional):
            An object that provides the parametrized form of a
            multiplicative function included in the model.
    """

    bitmask = RukiaBitMask()
    """
    Bitmask for Rukia results.
    """

    def __init__(self, sigma_model, mul_model=None, add_model=None):

        # Saved for convenience
        self._c = astropy.constants.c.to('km/s').value

        self.sigma = sigma_model
        self.nsp = self.sigma.np
        self.mul = mul_model
        self.nmp = 0 if self.mul is None else self.mul.np
        self.add = add_model
        self.nap = 0 if self.add is None else self.add.np
        self.np = 1 + self.nsp + self.nmp + self.nap

        # Model parameters
        self.par = numpy.ones(self.np, dtype=float)
        # Initialize all parameters as free
        self.free = numpy.ones(self.np, dtype=bool)
        self.nfree = numpy.sum(self.free)

        # Objects with data for the spectrum to fit
        self.wave_scaled = None
        self.wave = None
        self.dw = None
        self.flux = None
        self.err = None
        self.gpm = None
        self.sres = None
        # Objects with data for the template spectrum
        self.tpl_wave = None
        self.tpl_dw = None
        self.tpl_flux = None
        self.tpl_sres = None

        # Bitmask value for modeling results/input
        self.flags = self.bitmask.minimum_dtype()(0)

        # Range is initialized to None and then changed by the fit()
        # method
        # TODO: Issue warning if any of the models have rng defined?
        # This will overwrite them.
        self._reset_range(None)

    @staticmethod
    def _init_data(wave, flux, err, mask, sres):
        # Set the data to fit
        if flux.ndim > 1:
            raise ValueError('Flux vector must be 1D.')
        _mask = numpy.zeros(flux.shape, dtype=bool) if mask is None else mask.copy()
        if _mask.shape != flux.shape:
            raise ValueError('Mask and flux vectors do not have the same shape.')
        if isinstance(flux, numpy.ma.MaskedArray):
            _flux = flux.data.copy()
            _mask |= numpy.ma.getmaskarray(flux)
        else:
            _flux = flux.copy()
        if wave.shape != _flux.shape:
            raise ValueError('Wavelength and flux vectors do not have the same shape.')
        _dw = angstroms_per_pixel(wave, regular=False)
        _err = None
        if err is not None:
            if err.shape != _flux.shape:
                raise ValueError('Flux error and flux vectors do not have the same shape.')
            if isinstance(err, numpy.ma.MaskedArray):
                _err = err.data.copy()
                _mask |= numpy.ma.getmaskarray(err)
            else:
                _err = err.copy()
        _sres = None
        if sres is not None:
            if sres.shape != _flux.shape:
                raise ValueError('Spectral resolution and flux vectors do not have the same shape.')
            _sres = interpolate_masked_vector(sres) if isinstance(sres, numpy.ma.MaskedArray) \
                        else sres.copy()
        return _dw, _flux, _err, numpy.logical_not(_mask), _sres

    @property
    def has_template(self):
        return self.tpl_wave is not None and self.tpl_flux is not None

    def shifted_wave(self, wave, shift=None):
        if shift is None:
            shift = self.par[0]
        return wave/(1+shift/self._c)

    def model(self, par, wave=None, tpl_wave=None, tpl_flux=None, rng=None,
              pixel_sigma_bounds=[0.01,100]):
        r"""
        Construct the model spectrum.

        Args:
            par (`numpy.ndarray`_):

                List of model parameters. Shape must be
                :math:`(N_{\rm par},)` or the number of free
                parameters (see :attr:`free`).
            
            wave (`numpy.ndarray`_, optional):
                The wavelength vector on which to construct the
                model. If not provided, the wavelengths for the model
                must have already been defined for this instance,
                e.g., via a fit to a provided spectrum (see
                :func:`fit`).
            tpl_wave(`numpy.ndarray`_, optional):
                Wavelength vector for template spectrum. If provided,
                ``tpl_flux`` must also be provided. If not provided,
                the template spectrum must have already been defined
                for this instance, e.g., via a fit to a provided
                spectrum (see :func:`fit`).
            tpl_flux(`numpy.ndarray`_, optional):
                Flux vector for template spectrum. If provided,
                ``tpl_wave`` must also be provided. If not provided,
                the template spectrum must have already been defined
                for this instance, e.g., via a fit to a provided
                spectrum (see :func:`fit`).
            rng (`numpy.ndarray`_, optional):
                The wavelength range used to scale the functional
                forms used for the model parameters. If None, the
                range is assumed to be already defined or irrelevant
                to the construction of the model.

        Returns:
            `numpy.ndarray`_: The model spectrum.
        """
        if (tpl_wave is None or tpl_flux is None) and not self.has_template:
            raise ValueError('No template data available to construct model.')
        if (tpl_wave is None and tpl_flux is not None) \
                or (tpl_wave is not None and tpl_flux is None):
            raise ValueError('If providing tpl_wave or tpl_flux, must provide both.')
        if wave is None and self.wave is None:
            raise ValueError('No wavelength vector available for model.')
        if rng is not None:
            self._reset_range(rng)

        self._set_par(par)

        _wave_scaled = self.wave_scaled if wave is None \
                            else modeling.scaled_coordinates(wave, rng=self.rng)
        _wave = self.wave if wave is None else wave
        _tpl_wave = self.shifted_wave(self.tpl_wave if tpl_wave is None else tpl_wave)
        _tpl_wave_scaled = modeling.scaled_coordinates(_tpl_wave, rng=self.rng)
        _tpl_flux = self.tpl_flux if tpl_flux is None else tpl_flux

        # Sigma in angstroms
        sigma = self.sigma(_tpl_wave_scaled, a=self.sigma_par())
        # Convert it to pixels
        dw = angstroms_per_pixel(_tpl_wave, regular=False)
        sigma /= dw
        # Imposes bounds on the allowed sigma in pixels for the
        # convolution. The minimum is, by default, set to be the same
        # as what is used by ppxf.ppxf_util.gaussian_filter1d
        self.flags = 0
        if numpy.any(sigma <= pixel_sigma_bounds[0]):
            self.flags = self.bitmask.turn_on(self.flags, 'LOSIGMA')
        if numpy.any(sigma >= pixel_sigma_bounds[1]):
            self.flags = self.bitmask.turn_on(self.flags, 'HISIGMA')
        # Clip the sigma to the specified bounds
        sigma = numpy.clip(sigma, *pixel_sigma_bounds)

        # TODO: Should insist that spectra are regularly gridded.
        # Otherwise, the convertion to pixels below is not strictly
        # true if there are significant differences in the width of
        # adjacent pixels, relative to the width of the kernel.

        # Convolve the template spectrum. The
        # gaussian_filter1d algorithm results in the first and last
        # ``p`` pixels being set to 0.
        p = int(numpy.ceil(numpy.max(3*sigma)))
        model_flux = gaussian_filter1d(_tpl_flux, sigma)[p:-p]

        # Resample the convolved data to the output wavelength grid
        model_flux = Resample(model_flux, x=_tpl_wave[p:-p], newx=_wave).outy

        # Multiply by any polynomial
        if self.mul is not None:
            model_flux *= self.mul(_wave_scaled, a=self.mul_par())
        # Include an additive polynomial
        if self.add is not None:
            model_flux += self.add(_wave_scaled, a=self.add_par())
        return model_flux

    def _set_par(self, par):
        """
        Set the parameters by accounting for any fixed parameters.
        """
        if par.ndim != 1:
            raise ValueError('Parameter array must be a 1D vector.')
        if par.size == self.np:
            self.par = par.copy()
            return
        if par.size != self.nfree:
            raise ValueError('Must provide {0} or {1} parameters.'.format(self.np, self.nfree))
        self.par[self.free] = par.copy()

    def _reset_range(self, rng):
        self.rng = rng
        for f in [self.sigma, self.add, self.mul]:
            if f is None:
                continue
            if hasattr(f, 'reset_range'):
                f.reset_range(rng)
        if self.wave is not None:
            self.wave_scaled = modeling.scaled_coordinates(self.wave, rng=self.rng)

    def _sigma_par_slice(self):
        return slice(1,1+self.nsp)

    def sigma_par(self, par=None):
        if par is None:
            par = self.par
        return par[self._sigma_par_slice()]

    def _mul_par_slice(self):
        s = 1+self.nsp
        return slice(s,s+self.nmp)

    def mul_par(self, par=None):
        if par is None:
            par = self.par
        return par[self._mul_par_slice()]

    def _add_par_slice(self):
        s = 1+self.nsp+self.nmp
        return slice(s,s+self.nap)

    def add_par(self, par=None):
        if par is None:
            par = self.par
        return par[self._add_par_slice()]

    def _init_par(self, shift=0.0, fit_shift=True, sigma_p0=None, sigma_f=None, add_p0=None,
                  add_f=None, mul_p0=None, mul_f=None):
        # Check the input parameters
        if sigma_p0 is not None and len(sigma_p0) != self.nsp:
            raise ValueError('Incorrect number of parameters for the sigma model.')
        if sigma_f is not None and len(sigma_f) != self.nsp:
            raise ValueError('Incorrect number of parameter flags for the sigma model.')
        if sigma_p0 is None and sigma_f is not None:
            warnigns.warn('Did not provide sigma model parameters, but did provide sigma model '
                          'parameter fitting flags.  Parameters may get fixed to default guesses.')
        if add_p0 is not None and len(add_p0) != self.nap:
            raise ValueError('Incorrect number of parameters for the additive polynomial.')
        if add_f is not None and len(add_f) != self.nap:
            raise ValueError('Incorrect number of parameter flags for the additive polynomial.')
        if add_p0 is None and add_f is not None:
            warnigns.warn('Did not provide additive polynomial parameters, but did provide '
                          'additive polynomial parameter fitting flags.  Parameters may get '
                          'fixed to default guesses.')
        if mul_p0 is not None and len(mul_p0) != self.nmp:
            raise ValueError('Incorrect number of parameters for the multiplicative polynomial.')
        if mul_f is not None and len(mul_f) != self.nmp:
            raise ValueError('Incorrect number of parameter flags for the multiplicative '
                             'polynomial.')
        if mul_p0 is None and mul_f is not None:
            warnigns.warn('Did not provide multiplicative polynomial parameters, but did provide '
                          'multiplicative polynomial parameter fitting flags.  Parameters may get '
                          'fixed to default guesses.')

        # Initialize the parameter vector
        self.par[0] = shift
        self.free[0] = fit_shift

        # Sigma parameters:
        if self.nsp > 0:
            ps = self._sigma_par_slice()
            self.par[ps] = numpy.array([1.] + [0.]*(self.nsp-1)) if sigma_p0 is None else sigma_p0
            self.free[ps] = True if sigma_f is None else sigma_f

        # Multiplicative polynomial parameters:
        if self.nmp > 0:
            ps = self._mul_par_slice()
            self.par[ps] = numpy.array([1.] + [0.]*(self.nmp-1)) if mul_p0 is None else mul_p0
            self.free[ps] = True if mul_f is None else mul_f

        # Additive polynomial parameters:
        if self.nap > 0:
            ps = self._add_par_slice()
            self.par[ps] = numpy.array([1.] + [0.]*(self.nap-1)) if add_p0 is None else add_p0
            self.free[ps] = True if add_f is None else add_f

        self.nfree = numpy.sum(self.free)

    def fitting_range(self, wave, tpl_wave, shift=None, shift_range=None, rng=None):

        # Determine the spectral range to fit
        if wave[-1] < tpl_wave[0] or wave[0] > tpl_wave[-1]:
            raise ValueError('No overlapping spectral region between the spectrum to fit and '
                             'the template.')

        # 300 here is matched to the maximum sigma allowed of 100 in
        # gaussian_filter1d and how it maskes the edges of the
        # spectrum.
        tpl_wave_lim = [self.shifted_wave(tpl_wave[300], shift=shift if shift_range is None
                                                               else shift+shift_range[0]),
                        self.shifted_wave(tpl_wave[-300], shift=shift if shift_range is None
                                                                else shift+shift_range[1])]

        # Get the overlapping spectral range to fit
        # TODO: Alter this based on the range of allowed shifts
        _rng = numpy.array([max(wave[0], tpl_wave_lim[0]), min(wave[-1], tpl_wave_lim[-1])])

        if rng is None:
            return _rng

        if rng[0] < _rng[0] or rng[0] > _rng[1] or rng[1] > _rng[1] or rng[1] < _rng[0]:
            raise ValueError('Provided range is invalid.  Maximum range is: '
                             '{0}, but requested range is {1}'.format(_rng, rng))
        return rng

    def fit(self, wave, flux, tpl_wave, tpl_flux, err=None, mask=None, sres=None, tpl_sres=None,
            shift=0.0, fit_shift=True, shift_range=[-300.,300.], rejiter=None, rejsig=3.0,
            rejbox=101, sigma_p0=None, sigma_f=None, add_p0=None, add_f=None, mul_p0=None,
            mul_f=None, rng=None):
        r"""
        Fit the spectrum.

        If err is provided:
            - the fit figure-of-merit is chi-square
            - rejections are based on error-normalized residuals
        Otherwise:
            - the fit figure-of-merit is the fit RMS
            - rejections are based on un-normalized residuals.

        Args:
            shift (:obj:`float`, optional):
                The Doppler (multiplicative) shift between the input
                spectrum wavelengths and the template wavelengths;
                i.e., this is :math:`z` is the Doppler shift such
                that the best match in the spectra is achieved when
                the wavelengths of the spectrum to fit are
                
                .. math::
                    
                    \lambda = (1+z) \lambda_{\rm tpl}.

                If ``fit_shift`` is False, this shift is fixed during
                the fit; otherwise, it is included as a fitted model
                parameter and the provided value is used as an
                initial guess.
            fit_shift (:obj:`bool`, optional):
                In addition to the broadening, fit a Doppler (i.e.,
                multiplicative) shift between the template and target
                wavelengths.
        """
        self._init_par(shift=shift, fit_shift=fit_shift, sigma_p0=sigma_p0, sigma_f=sigma_f,
                       add_p0=add_p0, add_f=add_f, mul_p0=mul_p0, mul_f=mul_f)

        self.wave = wave
        self.tpl_wave = tpl_wave
        self._reset_range(self.fitting_range(wave, tpl_wave, shift, shift_range=shift_range,
                          rng=rng))

        indx = (self.wave > self.rng[0]) & (self.wave < self.rng[1])
        if numpy.sum(indx) != len(indx):
            warnings.warn('Limiting fit to {0} - {1}.'.format(*self.rng))

        # Setup the spectrum to fit
        self.wave_scaled = self.wave_scaled[indx]
        self.wave = self.wave[indx]
        self.dw, self.flux, self.err, self.gpm, self.sres \
                = self._init_data(wave[indx], flux[indx], None if err is None else err[indx],
                                  None if mask is None else mask[indx],
                                  None if sres is None else sres[indx])

        # Setup the template spectrum
        self.tpl_dw, self.tpl_flux, _, _, self.tpl_sres \
                = self._init_data(tpl_wave, tpl_flux, None, None, tpl_sres)

        lb = numpy.full(self.np, -numpy.inf, dtype=float)
        ub = numpy.full(self.np, numpy.inf, dtype=float)
        lb[0], ub[0] = shift + numpy.array(shift_range)

        # Set the figure-of-merit for the fit, perform the fit, and get
        # the parameter errors from the covariance matrix
        resid = modeling.FitResiduals(self.flux, self.model, gpm=self.gpm) if self.err is None \
                    else modeling.FitChiSquare(self.flux, self.err, self.model, gpm=self.gpm)

        # TODO: Change this for the sigma parameters too?
        diff_step = numpy.full(self.np, numpy.sqrt(numpy.finfo(float).eps), dtype=float)
        diff_step[0] = 0.1
        result = optimize.least_squares(resid, self.par[self.free], method='trf',
                                        diff_step=diff_step[self.free],
                                        bounds=(lb[self.free], ub[self.free])) #, verbose=2)

        if rejiter is None or rejiter == 0:
            self._set_par(result.x)
            return

        # Keep the initial mask
        init_gpm = self.gpm.copy()

        # Iteratively reject
        i = 0
        while i < rejiter or rejiter < 0:
            # Reject pixels
            rej = modeling.reject_residuals_1d(resid(self.par), lo=rejsig, hi=rejsig,
                                               boxcar=rejbox)
            if not numpy.any(rej):
                # None were rejected so we're done
                break
            # Mask the rejected values
            self.gpm[numpy.where(self.gpm)[0][rej]] = False
            # Refit, starting at the previous fit position
            result = optimize.least_squares(resid, self.par[self.free], method='trf',
                                            diff_step=diff_step[self.free],
                                            bounds=(lb[self.free], ub[self.free])) #, verbose=2)
            # Reset the best-fit parameters
            self._set_par(result.x)
            # Increment iteration
            i += 1







#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.97')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax


def mass_sigma(dapall, drpall):
    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    indx &= (dapall['STELLAR_SIGMA_1RE'] > 0)
    indx &= (dapall['NSA_ELPETRO_TH50_R'] > 0)
    indx &= (dapall['ADIST_Z'] > 0)

    rekpc = dapall['NSA_ELPETRO_TH50_R']*numpy.radians(1/3600)*1e3*dapall['ADIST_Z']

    font = { 'size' : 18 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    mass_lim = [7e7, 3e12]
    sigm_lim = [10,  800]

    inst = numpy.array([[mass_lim[0], 70],
                        [mass_lim[1], 70],
                        [mass_lim[1], sigm_lim[0]],
                        [mass_lim[0], sigm_lim[0]]])

    ax = init_ax(fig, [0.2, 0.20, 0.6, 0.6])
    ax.set_xlim(mass_lim)
    ax.set_ylim(sigm_lim)
    ax.set_xscale('log')
    ax.set_yscale('log')

    x = numpy.linspace(6, 14, 100)
#    y = x/4 - 0.25
#    ax.plot(numpy.power(10,x), numpy.power(10,y-1.5), color='0.8', zorder=0)
#    ax.plot(numpy.power(10,x), numpy.power(10,y-0.75), color='0.8', zorder=0)
#    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
#    ax.plot(numpy.power(10,x), numpy.power(10,y+0.75), color='0.8', zorder=0)

    y = x/2 - 2.6
    ax.plot(numpy.power(10,x), numpy.power(10,y-2), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y-1), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+1), color='0.8', zorder=0)

    ax.fill(inst[:,0], inst[:,1], color='0.1', alpha=0.1, zorder=1)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][indx]],
                    dapall['STELLAR_SIGMA_1RE'][indx], lw=0, marker='.', s=20, zorder=3,
                    c=rekpc[indx],
#                    vmin=0.1, vmax=18, cmap='inferno')
                    norm=colors.LogNorm(vmin=1, vmax=30), cmap='inferno')

    cax = fig.add_axes([0.63, 0.26, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1,10],
                         format=ticker.ScalarFormatter())
    cax.text(0.5, 3.0, r'$R_e$ [kpc]', ha='center', va='center', transform=cax.transAxes)
    

    ax.text(0.5, -0.15, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.2, 0.5, r'$\sigma_{\ast,e}$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')
    ax.text(0.51, 0.85, r'$\mathcal{M}_\ast\propto\ \sigma^2_{\ast,e}$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation=53, zorder=4)
    ax.text(0.99, 0.37, r'$\sigma_{\ast,e}\ \lesssim\ \sigma_{\rm inst}$',
            horizontalalignment='right', verticalalignment='center', transform=ax.transAxes,
            zorder=4)

    ofile = '../ms/figs/mass_sigma.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data

    drpall_file = '/Users/westfall/Work/MaNGA/testing/dr15/drpall-v2_4_3.fits'
    drpall = fits.open(drpall_file)[1].data

    mass_sigma(dapall, drpall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))





import sys
if sys.version > '3':
    long = int

import os
import gzip
import shutil
import logging
import numpy as np
import pdb

from astropy.io import fits
from astropy.io import ascii

from mangadap.util import sampling
from mangadap.util import fileio

## Need FWHM of resolution element in Ang
## Vacuum or air?


select_ages = np.arange(-1.2,1.2,0.1)
#allZ = np.array([0.001, 0.002, 0.003, 0.004, 0.006, 0.008, 0.010, 0.014, 0.020, 0.030, 0.040])
select_Z = np.array([0.008, 0.02, 0.03])

#bpass_fil = './v2.2.1_imf135_300/spectra-bin-imf135_300.z020.dat'
bpass_root = '/Users/rubin/Research/MaNGA/NaImcmc/Analysis/BPASS/v2.2.1_imf135_300/spectra-bin-imf135_300.z'
out_root = '/Users/rubin/Repo/manga/mangadap/data/spectral_templates/bpass/'

## Set up info common to all files
ncol = 52
col_arr = np.arange(1,ncol+1)
age_arr = 10.0**(6.0 + 0.1*(col_arr-2.0))
logage_arr = np.log10(age_arr / 1.0e9)

#pdb.set_trace()
for Z in select_Z:

    Z_str = "{:.3f}".format(Z)[2:]
    bpass_fil = bpass_root+Z_str+'.dat'

    data = ascii.read(bpass_fil)
    wave = data['col1']
    wvrange = (wave > 3600.0) & (wave < 10400.0)
    dlam = wave[wvrange] - np.roll(wave[wvrange],1)
    npix = len(dlam)

    crval = min(wave[wvrange])
    crpix = 1.0
    cdelt = np.median(dlam)
    #test_wv = (np.arange(1.0,npix+1) - crpix)*cdelt + crval

    for age in select_ages:

        ind = ((logage_arr>age-0.001) & (logage_arr<age+0.001))
        #print(logage_arr[ind])
        if(len(logage_arr[ind])>0):
            colname = 'col'+str(col_arr[ind][0])
            flux = data[colname]
            out_flux = flux[wvrange]

            age_str = "{:.2f}".format(age)
            out_fil = out_root + 'bin-imf135_300-z'+Z_str+'-T'+age_str+'.fits'
            print(out_fil)
   
            ## Maybe can just use fileio.writefits_1dspec
            fileio.writefits_1dspec(out_fil, crval, cdelt, out_flux, clobber=True)

            #step = sampling.spectral_coordinate_step(wave)

            #pdb.set_trace()



#!/usr/bin/env python3

import numpy
from os import remove
import os.path
from time import clock

from mangadap.drpfile import drpfile
from mangadap.proc.TemplateLibrary import TemplateLibrary
from matplotlib import pyplot
from mangadap.util.par import TemplateLibraryParSet
from mangadap.util.defaults import default_dap_source

from mangadap.util.fileio import readfits_1dspec
from mangadap.util.instrument import spectral_resolution, spectrum_velocity_scale

#-----------------------------------------------------------------------------

def overplot_test():
    drpf = drpfile(7495, 12703, 'CUBE')

    tpl_lib = TemplateLibrary('M11-MILES', process=False)
    unmasked = numpy.invert( tpl_lib.tplbm.flagged(tpl_lib.hdu['MASK'].data) )

    print(tpl_lib.file_list)
    print(tpl_lib.ntpl)

    i = 0

    wave_0, flux_0 = readfits_1dspec(tpl_lib.file_list[i])

    wave_t = tpl_lib.hdu['WAVE'].data[i,unmasked[i,:]]
    flux_t = tpl_lib.hdu['FLUX'].data[i,unmasked[i,:]]

    print(wave_0.shape)
    print(wave_t.shape)
    print(flux_0.shape)
    print(flux_t.shape)

    pyplot.plot(tpl_lib.hdu['WAVE'].data[i,unmasked[i,:]],tpl_lib.hdu['FLUX'].data[i,unmasked[i,:]])
    pyplot.plot(wave_0, flux_0)
    pyplot.show()

    pyplot.plot(wave_0, flux_0-flux_t)
    pyplot.show()


def run_test():

#    drpf = drpfile(7495, 12703, 'CUBE')
    drpf = drpfile(7495, 12703, 'CUBE', read=True)

    sres = spectral_resolution(drpf.hdu['WAVE'].data, drpf.hdu['SPECRES'].data, log10=True)
    velscale = spectrum_velocity_scale(drpf.hdu['WAVE'].data, log10=True)

    print(drpf.file_path())

#    tpl_lib = TemplateLibrary('M11-MILES', drpf=drpf, directory_path='.', force=True)
    tpl_lib = TemplateLibrary('STELIB', velscale=velscale, sres=sres, directory_path='.',
                              processed_file='test.fits', force=True)
#    tpl_lib = TemplateLibrary('STELIB', process=False)

#    dapsrc = default_dap_source()
#
#    # M11-MARCS
#    template_libraries = TemplateLibraryParSet(key='M11-MARCS',
#                                file_search=dapsrc+'/external/templates/m11_marcs/*_s.fits',
#                                # TODO: This is the resolution in the header of the files, is it
#                                # right?
#                                fwhm=2.73, in_vacuum=False,
#                                wave_limit=numpy.array([ None, None ]),
#                                lower_flux_limit=None)
#    tpl_lib = TemplateLibrary('MILES-AVG', tpllib_list=template_libraries, process=False)
    
    print(tpl_lib.hdu['FLUX'].shape)
    nspec = tpl_lib.hdu['FLUX'].shape[0]

    unmasked = numpy.invert( tpl_lib.tplbm.flagged(tpl_lib.hdu['MASK'].data, flag=['NO_DATA','WAVE_INVALID','FLUX_INVALID']) )

#    edges = numpy.linspace(0,985,num=11,dtype=int)
#    for j in range(0,10):
#        for i in range(edges[j],edges[j+1]):
#        print(tpl_lib.file_list[i])

    for i in range(0,nspec):
#    for i in range(969,970):
#        print(tpl_lib.file_list[i])
#        pyplot.plot(tpl_lib.hdu['WAVE'].data[i,unmasked[i,:]], tpl_lib.hdu['FLUX'].data[i,unmasked[i,:]])
        x = tpl_lib.hdu['WAVE'].data[:]
#        x = tpl_lib.hdu['WAVE'].data[i,:]
        y = tpl_lib.hdu['FLUX'].data[i,:]
        pyplot.plot(x[unmasked[i,:]], y[unmasked[i,:]], ms=5, linestyle='none', marker='.', markeredgecolor=None, color='black')
        pyplot.plot(x[numpy.invert(unmasked[i,:])], y[numpy.invert(unmasked[i,:])], ms=10, linestyle='none', marker='.', markeredgecolor=None, color='red')
        df = (numpy.amax(y)-numpy.amin(y))*1.2
        f0 = (numpy.amax(y)+numpy.amin(y)-df)/2.0
        pyplot.ylim([f0, f0+df])
#        pyplot.plot(tpl_lib.hdu['WAVE'].data[unmasked[i,:]], tpl_lib.hdu['FLUX'].data[i,unmasked[i,:]])
        pyplot.show()

#    tpl_lib.hdu.info()
#    print(tpl_lib.hdu['WAVE'].shape)
#    print(tpl_lib.hdu['FLUX'].shape)
#    pyplot.plot(tpl_lib.hdu['WAVE'].data[101,:], tpl_lib.hdu['FLUX'].data[101,:])
#    pyplot.show()


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = clock()

    #overplot_test()
    run_test()

    print('Elapsed time: {0} seconds'.format(clock() - t))






#!/usr/bin/env python3

"""
Dynamically build the rst documentation of the bitmasks.
"""

import os
import time

#-----------------------------------------------------------------------------

def write_datatable(datatable_class, opath, class_link=True):
    ofile = os.path.join(opath, '{0}.rst'.format(datatable_class.__name__.lower()))
    lines = datatable_class().to_rst_table(header=False, class_link=class_link)
    with open(ofile, 'w') as f:
        f.write('\n'.join(lines))

if __name__ == '__main__':
    t = time.perf_counter()

    from pkg_resources import resource_filename
    root = os.path.dirname(resource_filename('mangadap', ''))
    path = os.path.join(root, 'docs', 'tables')
    if not os.path.isdir(path):
        os.makedirs(path)

    # Tables to write:
    #
    from mangadap.proc.reductionassessments import ReductionAssessmentDataTable
    write_datatable(ReductionAssessmentDataTable, path, class_link=False)
    from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectraDataTable
    write_datatable(SpatiallyBinnedSpectraDataTable, path, class_link=False)
    from mangadap.proc.spectralfitting import StellarKinematicsFitDataTable
    write_datatable(StellarKinematicsFitDataTable, path, class_link=False)
    from mangadap.par.emissionmomentsdb import EmissionMomentsDefinitionTable
    write_datatable(EmissionMomentsDefinitionTable, path, class_link=False)
    from mangadap.proc.emissionlinemoments import EmissionLineMomentsDataTable
    write_datatable(EmissionLineMomentsDataTable, path, class_link=False)
    from mangadap.par.emissionlinedb import EmissionLineDefinitionTable
    write_datatable(EmissionLineDefinitionTable, path, class_link=False)
    from mangadap.proc.spectralfitting import EmissionLineFitDataTable
    write_datatable(EmissionLineFitDataTable, path, class_link=False)
    from mangadap.proc.spectralindices import SpectralIndicesDefinitionTable
    write_datatable(SpectralIndicesDefinitionTable, path, class_link=False)
    from mangadap.proc.spectralindices import SpectralIndicesDataTable
    write_datatable(SpectralIndicesDataTable, path, class_link=False)

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






"""
Class that performs basic assessments of the reduced data products needed for
handling of the data by the DAP.  These assessments need only be done once per
data file.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import inspect
from pathlib import Path
import warnings
import logging

from IPython import embed

import numpy

from astropy.io import fits

from pydl.goddard.astro import airtovac

#from ..datacube import DataCube
from ..par.parset import KeywordParSet

from ..config import defaults

from ..util.datatable import DataTable
from ..util.fitsutil import DAPFitsUtil
from ..util.covariance import Covariance
from ..util.geometry import SemiMajorAxisCoo
from ..util.fileio import create_symlink
from ..util.log import log_output

# TODO: Add center coordinates
#            center_coo (:obj:`tuple`, optional):
#                A two-tuple with the coordinates in right-ascension
#                and declination for the coordinate-frame origin. If
#                None, no offset is performed.

# TODO: Fix the fact that if the correlation matrix is taken directly
# from the datacube that the wavelength and channel won't be defined.
# TODO: Change 'waverange' to 'pixelmask'?
class ReductionAssessmentDef(KeywordParSet):
    """
    Define how to assess the reduced data.

    The defined parameters are:

    .. include:: ../tables/reductionassessmentdef.rst

    """
    def __init__(self, key='SNRG', waverange=None, response_func_file='gunn_2001_g_response.db',
                 in_vacuum=True, covariance=True, minimum_frac=0.8, overwrite=False):

        # Use the signature to get the parameters and the default values
        sig = inspect.signature(self.__class__)
        pars = list(sig.parameters.keys())
        defaults = [sig.parameters[key].default for key in pars]

        # Remaining definitions done by hand
        ar_like = [ numpy.ndarray, list ]
        in_fl = [ int, float ]

        values = [key, waverange, response_func_file, in_vacuum, covariance, minimum_frac,
                  overwrite]
        dtypes = [str, ar_like, str, bool, bool, in_fl, bool]
        descr = ['Keyword to distinguish the assessment method.',
                 'A two-element vector with the starting and ending wavelength within which to ' \
                    'calculate the signal-to-noise.  Mutually exclusive with response_func_file',
                 'The name of a file that defines a response function to use for the S/N ' \
                    'calculation.  Must be a local file or distributed with the DAP source code.' \
                    '  Expected to have two columns, with the wavelength and response ' \
                    'efficiency.  Mutually exclusive with waverange.',
                 'Boolean indicating that the wavelengths provided either using waverange or ' \
                    'response_func_file are in vacuum (not air).',
                 'Provide the fiducial spatial covariance.  If this is False, no spatial ' \
                     'covariance will be available for calculations that use the results of the ' \
                     'reduction assessments.  If True and a covariance matrix is available ' \
                     'directly from the datacube, it will be used.  If True and the datacube ' \
                     'does not already provide a computed covariance matrix, one is calculated ' \
                     'using :func:`~mangadap.datacube.datacube.DataCube.covariance_matrix` ' \
                     'method.  **WARNING**: If either of these fail, the DAP will issue a ' \
                     'warning and continue assuming no spatial covariance.',
                 'Minimum fraction of unmasked pixels in a spectrum required for inclusion in ' \
                     'the spatial covariance calculation.  Note this should match the value ' \
                     'used for the spatial-binning module.',
                 'If the output file already exists, redo all the calculations and overwrite it.']

        super().__init__(pars, values=values, defaults=defaults, dtypes=dtypes, descr=descr)
        self.response = None
        self._validate()

    def _validate(self):
        """
        Validate the parameters.
        """
        if self['waverange'] is not None and self['response_func_file'] is not None:
            warnings.warn('You have defined both a wavelength range and a response function for '
                          'the reduction assessments; the latter takes precedence, the former is '
                          'ignored!')
            self['waverange'] = None
        if self['waverange'] is None and self['response_func_file'] is None:
            raise ValueError('Reduction assessment method undefined.  Must provide either '
                             '\'waverange\' or \'response_func_file\'.')

        if self['waverange'] is not None:
            if len(self['waverange']) != 2:
                raise ValueError('Defined wavelength range must be a two-element list/array.')
            # Wavelength range was defined
            if not self['in_vacuum']:
                # Convert it to vacuum
                self['waverange'] = airtovac(self['waverange'])
            return

        response_file = Path(self['response_func_file']).resolve()
        if not response_file.exists():
            # File not found, so try to find it in the configuration directory
            response_file = defaults.dap_data_root() / 'filter_response' \
                                / self['response_func_file']
        if not response_file.exists():
            raise FileNotFoundError(f'Could not find {self["response_func_file"]} '
                                    'in the local directory or the DAP source distribution.')

        self.response = numpy.genfromtxt(str(response_file))[:,:2]
        if not self['in_vacuum']:
            self.response[:,0] = airtovac(self.response[:,0])


class ReductionAssessmentDataTable(DataTable):
    """
    Primary data table with the reduction assessments.

    The data table includes:

    .. include:: ../tables/reductionassessmentdatatable.rst

    Args:
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use :func:`init` to initialize
            the data array after instantiation.
    """
    def __init__(self, shape=None):
        # NOTE: This should require python 3.7 to make sure that this
        # is an "ordered" dictionary.
        datamodel = dict(SPAT_INDX=dict(typ=int, shape=(2,),
                            descr='The 2D spatial array indices associated with each measurement. '
                                'See :attr:`~mangadap.datacube.datacube.DataCube.spatial_index`.'),
                         SKY_COO=dict(typ=float, shape=(2,),
                            descr='On-sky X and Y coordinates.  Coordinates are sky-right '
                                  'offsets from the object center; i.e., positive X is along the '
                                  'direction of positive right ascension. See '
                            ':class:`mangadap.datacube.datacube.DataCube.mean_sky_coordinates`.'),
                         ELL_COO=dict(typ=float, shape=(2,),
                                      descr='Elliptical (semi-major axis) radius and azimuth '
                                            'angle from N through E with respect to the '
                                            'photometric position angle; based on the provided '
                                            'ellipticity parameters.  See '
                                            ':class:`mangadap.util.geometry.SemiMajorAxisCoo`.'),
                         FGOODPIX=dict(typ=float, shape=None,
                                       descr='Fraction of good pixels in each spectrum.'),
#                         MINEQMAX=dict(typ=bool, shape=None,
#                                       descr='Flag that min(flux) = max(flux) in the spectrum; '
#                                             'i.e., the spaxel has no data.'),
                         SIGNAL=dict(typ=float, shape=None, descr='Per pixel mean flux'),
                         VARIANCE=dict(typ=float, shape=None, descr='Per pixel mean variance'),
                         SNR=dict(typ=float, shape=None, descr='Per pixel mean S/N'))

        keys = list(datamodel.keys())
        super(ReductionAssessmentDataTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               descr=[datamodel[k]['descr'] for k in keys],
                               shape=shape)

class ReductionAssessment:
    r"""
    Perform basic assessments of the datacube.

    These assessments are needed for handling of the data by the DAP
    and only to be done once per datacube.

    See :func:`compute` for the provided data.

    Args:
        method (:class:`~mangadap.proc.reductionassessments.ReductionAssessmentDef`):
            Object that defines the main parameters used for the assessment
            method.
        cube (:class:`mangadap.datacube.datacube.DataCube`):
            Datacube to analyze.
        pa (:obj:`float`, optional):
            On-sky position angle of the major axis used to calculate
            elliptical, semi-major-axis coordinates, defined as the
            angle from North through East and denoted :math:`\phi_0`.
        ell (:obj:`float`, optional):
            Ellipticity defined as :math:`\varepsilon=1-b/a`, based
            on the semi-minor to semi-major axis ratio (:math:`b/a`)
            of the isophotal ellipse used to calculate elliptical,
            semi-major-axis coordinates.
        reff (:obj:`float`, optional):
            Effective radius of the galaxy.  If None, set to 1.
        output_path (:obj:`str`, `Path`_, optional):
            The path for the output file.  If None, the current working
            directory is used.
        output_file (:obj:`str`, optional):
            The name of the output "reference" file. The full path of the output
            file will be :attr:`directory_path`/:attr:`output_file`.  If None,
            the default is to combine ``cube.output_root`` and the
            ``method_key``.  See
            :func:`~mangadap.proc.reductionassessments.ReductionAssessments.default_paths`.
        hardcopy (:obj:`bool`, optional):
            Flag to write the data to a fits file.
        symlink_dir (:obj:`str`, optional):
            Create a symlink to the file in this directory.
        checksum (:obj:`bool`, optional):
            Compare the checksum when reading the data.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects to log progress;
            ignored if quiet=True. Logging is done using
            :func:`mangadap.util.log.log_output`. Default is no
            logging.
        quiet (:obj:`bool`, optional):
            Suppress terminal and logging output.

    Attributes:
        method (:class:`ReductionAssessmentDef`):
            Parameters defining the method to use for the reduction
            assessments.
        cube (:class:`mangadap.datacube.datacube.DataCube`):
            Datacube to analyze.
        pa (:obj:`float`):
            On-sky position angle of the major axis used to calculate
            elliptical, semi-major-axis coordinates, defined as the
            angle from North through East and denoted :math:`\phi_0`.
        ell (:obj:`float`):
            Ellipticity defined as :math:`\varepsilon=1-b/a`, based
            on the semi-minor to semi-major axis ratio (:math:`b/a`)
            of the isophotal ellipse used to calculate elliptical,
            semi-major-axis coordinates.
        reff (:obj:`float`):
            Effective radius of the galaxy.
        output_path (:obj:`str`):
            The exact path for the output file.
        output_file (:obj:`str`):
            The name of the file for the computed assessments. The
            full path of the output file will be
            :attr:`directory_path`/:attr:`output_file`.
        hardcopy (:obj:`bool`):
            Flag to write the data to a fits file.
        symlink_dir (:obj:`str`):
            Create a symlink to the file in this directory.
        hdu (`astropy.io.fits.HDUList`_):
            HDUList with the data, with columns as described above.
        correlation (:class:`mangadap.util.covariance.Covariance`):
            If requested, provides the spatial correlation matrix for
            the flux at a fiducial wavelength in the datacube. See
            :attr:`covar_wave` and :attr:`covar_channel`.
        covar_wave (:obj:`float`):
            The wavelength in angstroms at which the correlation
            matrix (:attr:`correlation`) was calculated, if
            available. Can be None if the covariance matrix is pulled
            directly from :attr:`cube`.
        covar_channel (:obj:`int`):
            The wavelength channel at which the correlation matrix
            (:attr:`correlation`) was calculated, if available.
            Can be None if the covariance matrix is pulled
            directly from :attr:`cube`.
        loggers (:obj:`list`):
            List of `logging.Logger`_ objects to log progress;
            ignored if quiet=True. Logging is done using
            :func:`mangadap.util.log.log_output`.
        quiet (:obj:`bool`):
            Suppress terminal and logging output.

    """
    def __init__(self, method, cube, pa=None, ell=None, reff=None, output_path=None,
                 output_file=None, hardcopy=True, symlink_dir=None, checksum=False, loggers=None,
                 quiet=False):
                 
        self.loggers = None
        self.quiet = False

        # Define the method properties
#        self.method = None
#        self._define_method(method_key, method_list=method_list)
        self.method = method
        if not isinstance(self.method, ReductionAssessmentDef):
            raise TypeError('Method must have type ReductionAssessmentDef.')

        # Set in compute via _set_paths
        self.cube = None

        # Define the output directory and file.  Set by default_paths() in compute()
        self.directory_path = None
        self.output_file = None
        self.hardcopy = None
        self.symlink_dir = None

        # Initialize the objects used in the assessments
        self.pa = None
        self.ell = None
        self.reff = None
        self.hdu = None
        self.checksum = checksum
        self.correlation = None
        self.covar_wave = None
        self.covar_channel = None

        # Run the assessments of the DRP file
        self.compute(cube, pa=pa, ell=ell, reff=reff, output_path=output_path,
                     output_file=output_file, hardcopy=hardcopy, symlink_dir=symlink_dir,
                     loggers=loggers, quiet=quiet)

    def __getitem__(self, key):
        return self.hdu[key]

    def _initialize_primary_header(self, hdr=None):
        """
        Constuct the primary header for the reference file.

        Args:
            hdr (`astropy.io.fits.Header`_, optional):
                Input base header for added keywords. If None, uses
                the :attr:`cube` header (if there is one) and then
                cleans the header using
                :func:`mangadap.util.fitsutil.DAPFitsUtil.clean_dap_primary_header`.

        Returns:
             `astropy.io.fits.Header`_: Initialized header object.
        """
        # Copy the from the DRP and clean it
        if hdr is None:
            hdr = self.cube.prihdr.copy()
            hdr = DAPFitsUtil.clean_dap_primary_header(hdr)
        hdr['AUTHOR'] = 'Kyle B. Westfall <westfall@ucolick.org>'
        hdr['RDXQAKEY'] = (self.method['key'], 'Method keyword')
        if self.pa is not None:
            hdr['ECOOPA'] = (self.pa, 'Position angle for ellip. coo')
        if self.ell is not None:
            hdr['ECOOELL'] = (self.ell, 'Ellipticity (1-b/a) for ellip. coo')
        if self.reff is not None:
            hdr['ECOOREFF'] = (self.reff, 'Effective radius')
        if self.method['covariance']:
            hdr['BBWAVE'] = ('None' if self.covar_wave is None else self.covar_wave,
                             'Covariance channel wavelength')
            hdr['BBINDEX'] = ('None' if self.covar_channel is None else self.covar_channel,
                              'Covariance channel index')
        return hdr

    @staticmethod
    def default_paths(cube, method_key, output_path=None, output_file=None):
        """
        Set the default directory and file name for the output file.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to analyze.
            method_key (:obj:`str`):
                Keyword designating the method used for the reduction
                assessments.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the output
                file will be :attr:`directory_path`/:attr:`output_file`.  If None,
                the default is to combine ``cube.output_root`` and the
                ``method_key``.

        Returns:
            :obj:`tuple`: Returns a `Path`_ with the output directory and a
            :obj:`str` with the output file name.
        """
        directory_path = Path('.').resolve() if output_path is None \
                                else Path(output_path).resolve()
        _output_file = f'{cube.output_root}-{method_key}.fits.gz' if output_file is None \
                                else output_file
        return directory_path, _output_file

    def file_name(self):
        """Return the name of the output file."""
        return self.output_file

    def file_path(self):
        """Return a `Path`_ object with the full path to the output file."""
        if self.directory_path is None or self.output_file is None:
            return None
        return self.directory_path / self.output_file

    def info(self):
        return self.hdu.info()

    def compute(self, cube, pa=None, ell=None, reff=None, output_path=None, output_file=None, hardcopy=True,
                symlink_dir=None, overwrite=None, loggers=None, quiet=False):
        r"""
        Compute and output the main data products.  The list of HDUs
        are:

            - ``PRIMARY`` : Empty apart from the header information.
            - ``SPECTRUM`` : Extension with the main, per-spectrum
              measurements; see :class:`ReductionAssessmentDataTable`.
            - ``CORREL`` : The correlation matrix between the ``SIGNAL``
              measurements provided in the ``SPECTRUM`` extension.  The
              format of this extension is identical to the nominal
              output of the :class:`mangadap.util.covariance.Covariance`
              object; see
              :func:`mangadap.util.covariance.Covariance.write`.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to analyze.
            pa (:obj:`float`, optional):
                On-sky position angle of the major axis used to
                calculate elliptical, semi-major-axis coordinates,
                defined as the angle from North through East and
                denoted :math:`\phi_0`.
            ell (:obj:`float`, optional):
                Ellipticity defined as :math:`\varepsilon=1-b/a`,
                based on the semi-minor to semi-major axis ratio
                (:math:`b/a`) of the isophotal ellipse used to
                calculate elliptical, semi-major-axis coordinates.
            reff (:obj:`float`, optional):
                Effective radius of the galaxy.  If None, set to 1.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                ``method_key``.  See
                :func:`~mangadap.proc.reductionassessments.ReductionAssessments.default_paths`.
            hardcopy (:obj:`bool`, optional):
                Flag to write the data to a fits file.
            symlink_dir (:obj:`str`, optional):
                Create a symlink to the file in this directory.
            overwrite (:obj:`bool`, optional):
                If the output file already exists, this will force the
                assessments to be redone and the output file to be
                overwritten.  If None, uses the value in :attr:`method`.
            loggers (:obj:`list`, optional):
                List of `logging.Logger`_ objects to log progress;
                ignored if quiet=True. Logging is done using
                :func:`mangadap.util.log.log_output`. Default is no
                logging.
            quiet (:obj:`bool`, optional):
                Suppress terminal and logging output.

        Raises:
            TypeError:
                Raised if the provided ``cube`` is not a
                :class:`mangadap.datacube.datacube.DataCube`.
            ValueError:
                Raised if the output path cannot be defined.
        """

        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet

        # Check the input cube
#        if not isinstance(cube, DataCube):
#            raise TypeError('Must provide a valid DataCube object!')
        self.cube = cube

        # Test if the RSS file exists; cannot compute covariance if not
        if self.method['covariance'] and not self.cube.can_compute_covariance:
            warnings.warn('Datacube does not provide a covariance matrix and could not load RSS '
                          'counterpart for the covariance computation.  Continuing by assuming '
                          '**no** covariance calculations.')
            self.method['covariance'] = False

        # Reset the output paths if necessary
        self.directory_path, self.output_file \
                = ReductionAssessment.default_paths(cube, self.method['key'],
                                                    output_path=output_path,
                                                    output_file=output_file)

        # Check that the path for or to the file is defined
        ofile = self.file_path()
        if ofile is None:
            raise ValueError('File path for output file is undefined!')

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, f'{"REDUCTION ASSESSMENT COMPUTATIONS":^50}')
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, f'Output path: {self.directory_path}')
            log_output(self.loggers, 1, logging.INFO, f'Output file: {self.output_file}')

        _overwrite = self.method['overwrite'] if overwrite is None else overwrite

        # If the file already exists, and not overwriting, just read the file
        self.symlink_dir = symlink_dir
        if ofile.exists() and not _overwrite:
            self.hardcopy = True
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Reading exiting file')

            # Read the existing output file
            self.hdu = DAPFitsUtil.read(str(ofile), checksum=self.checksum)

            # Read the header data
            try:
                self.pa = self.hdu['PRIMARY'].header['ECOOPA']
            except:
                if not self.quiet:
                    warnings.warn('Unable to read position angle from file header!')
                self.pa = None
            if not self.quiet and pa is not None and self.pa != pa:
                warnings.warn('Provided position angle different from available file; set ' \
                              'overwrite=True to overwrite.')

            try:
                self.ell = self.hdu['PRIMARY'].header['ECOOELL']
            except:
                if not self.quiet:
                    warnings.warn('Unable to read ellipticity from file header!')
                self.ell = None
            if not self.quiet and ell is not None and self.ell != ell:
                warnings.warn('Provided ellipticity different from available file; set ' \
                              'overwrite=True to overwrite.')

            try:
                self.reff = self.hdu['PRIMARY'].header['ECOOREFF']
            except:
                if not self.quiet:
                    warnings.warn('Unable to read effective radius from file header!')
                self.reff = None
            if not self.quiet and reff is not None and self.reff != reff:
                warnings.warn('Provided effective radius different from available file; set ' \
                              'overwrite=True to overwrite.')

            if self.method['covariance']:
                # Construct the correlation matrix with the appropriate
                # variance
                self.correlation = Covariance.from_fits(self.hdu, ivar_ext=None,
                                                        correlation=True).apply_new_variance(
                                                            self.hdu['SPECTRUM'].data['VARIANCE'])
                self.covar_wave = self.hdu['PRIMARY'].header['BBWAVE']
                if isinstance(self.covar_wave, str):
                    self.covar_wave = eval(self.covar_wave)
                self.covar_channel = self.hdu['PRIMARY'].header['BBINDEX']
                if isinstance(self.covar_channel, str):
                    self.covar_channel = eval(self.covar_channel)
            else:
                self.correlation = None
                self.covar_wave = None
                self.covar_channel = None

            # Make sure the symlink exists
            if self.symlink_dir is not None:
                create_symlink(str(ofile), self.symlink_dir, overwrite=_overwrite)
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, '-'*50)
            return

        # (Re)Initialize some of the attributes
        if pa is not None:
            self.pa = pa
        if ell is not None:
            self.ell = ell
        if reff is not None:
            self.reff = reff

        # Initialize the record array for the SPECTRUM extension
#        spectrum_data = init_record_array(self.cube.nspec, self._per_spectrum_dtype())
        spectrum_data = ReductionAssessmentDataTable(shape=self.cube.nspec)
        spectrum_data['SPAT_INDX'] = numpy.asarray(tuple(self.cube.spatial_index))
        spectrum_data['SKY_COO'][:,0], spectrum_data['SKY_COO'][:,1] \
                = (x.ravel() for x in self.cube.mean_sky_coordinates())

        coord_conversion = SemiMajorAxisCoo(xc=0, yc=0, rot=0, pa=self.pa, ell=self.ell)
        spectrum_data['ELL_COO'][:,0], spectrum_data['ELL_COO'][:,1] \
                = coord_conversion.polar(spectrum_data['SKY_COO'][:,0],
                                         spectrum_data['SKY_COO'][:,1])

        # Get the number of good wavelength channels in each spaxel
        # TODO: Move this into DataCube?
        flags = self.cube.do_not_use_flags()
        flux = self.cube.copy_to_masked_array(flag=flags)
        spectrum_data['FGOODPIX'] = numpy.sum(numpy.logical_not(numpy.ma.getmaskarray(flux)),
                                              axis=1) / flux.shape[1]

        # Flag spaxels with a flux that doesn't vary
        # TODO: Is this superfluous now?
#        frange = numpy.ma.amax(flux, axis=1)-numpy.ma.amin(flux, axis=1)
#        spectrum_data['MINEQMAX'] = (numpy.logical_not(numpy.ma.getmaskarray(frange))) \
#                                        & (numpy.ma.absolute(frange) < 1e-10)

        # Set the wavelength range in the statistics. If a response
        # function is used, only use the wavelength region where the
        # response function is provided.
        waverange = [self.cube.wave[0], self.cube.wave[-1]] \
                    if self.method['waverange'] is None and self.method.response is None \
                    else (self.method['waverange'] if self.method['waverange'] is not None else
                        [ self.method.response[0,0], self.method.response[-1,0] ])
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Wavelength range for S and N calculation: {0:.1f} -- {1:.1f}'.format(
                                *waverange))

        # If calculating covariance, make sure that all the pixels in
        # the channel used and selected for analysis (at minimum based
        # on fraction of valid pixels in the spectrum - FGOODPIX) have
        # covariance measurements. This is done by first assuming the
        # channel at the center of the wavelength range and then
        # flipping back and forth across this channel until a valid
        # channel is found or you reach the limits of the defined
        # wavelength range. In the latter case an exception is raised!
        if self.method['covariance'] and self.cube.covar is not None:
            self.correlation = self.cube.covar.copy()
            self.correlation.to_correlation()
            self.covar_wave = None
            self.covar_channel = None
        elif self.method['covariance'] and self.cube.covar is None:
            self.covar_wave = self.cube.central_wavelength(waverange=self.method['waverange'],
                                                           response_func=self.method.response,
                                                           flag=flags)
            self.covar_channel = numpy.argsort(numpy.absolute(self.cube.wave-self.covar_wave))[0]
            goodfrac = spectrum_data['FGOODPIX'].reshape(self.cube.spatial_shape) \
                            > self.method['minimum_frac']
            off = 1
            sign = 1
            while self.covar_wave > waverange[0] and self.covar_wave < waverange[1] \
                    and numpy.sum(goodfrac & numpy.logical_not(
                                    self.cube.ivar[:,:,self.covar_channel] > 0)) > 0:
                self.covar_channel += sign*off
                self.covar_wave = self.cube.wave[self.covar_channel]
                sign *= -1
                off += 1

            if numpy.sum(goodfrac 
                         & numpy.logical_not(self.cube.ivar[:,:,self.covar_channel] > 0)) > 0:
                raise ValueError('Unable to find wavelength channel within fully valid data.')
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO,
                           'Covariance measured at wavelength: {0} (channel {1})'.format(
                                self.covar_wave, self.covar_channel))
            # Calculate the covariance matrix
            self.correlation = self.cube.covariance_matrix(self.covar_channel)
            self.correlation.to_correlation()
        else:
            self.correlation = None
            self.covar_wave = None
            self.covar_channel = None

        # Determine the statistics of the flux in the specified
        # wavelength range
        spectrum_data['SIGNAL'], spectrum_data['VARIANCE'], spectrum_data['SNR'] \
                = (x.ravel() for x in 
                        self.cube.flux_stats(waverange=self.method['waverange'],
                                             response_func=self.method.response, flag=flags))

        # Force the covariance matrix to use the mean variance instead
        # of the variance in the single, specified channel.
        if self.method['covariance']:
            self.correlation = self.correlation.apply_new_variance(spectrum_data['VARIANCE'])

        # Construct header
        hdr = self._initialize_primary_header()

        # Get the covariance hdu
        if self.method['covariance']:
            hdr, ivar_hdu, covar_hdu = self.correlation.output_hdus(reshape=True, hdr=hdr)

        # Get the main extension columns and construct the HDUList
        self.hdu = fits.HDUList([fits.PrimaryHDU(header=hdr),
                                 spectrum_data.to_hdu(name='SPECTRUM')])

        # Add the covariance information
        if self.method['covariance']:
            self.hdu += [ covar_hdu ]

        # Write the file
        if not self.directory_path.exists():
            self.directory_path.mkdir(parents=True)
        self.hardcopy = hardcopy
        if self.hardcopy:
            DAPFitsUtil.write(self.hdu, str(ofile), overwrite=_overwrite, checksum=True,
                              symlink_dir=self.symlink_dir, loggers=self.loggers)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)






from IPython import embed

import numpy
from scipy import special

from astropy.io import fits

from mangadap.util.constants import DAPConstants
from mangadap.tests.util import data_test_file
from mangadap.proc.rukia import Rukia, LegendrePolynomial


def pixelated_gaussian(x, c=0.0, s=1.0):
    """
    x must be linearly sampled

    x, c, and s must be in the same units.
    """
    n = numpy.sqrt(2.)*s
    d = numpy.asarray(x)-c
    dx = numpy.mean(numpy.diff(x))
    return (special.erf((d+dx/2.)/n) - special.erf((d-dx/2.)/n))/2.


def read_xsl(ifile, minimum_useful=True):
    if ifile is None:
        return None, None, None
    with fits.open(ifile) as hdu:
        wave = numpy.ma.MaskedArray(hdu['SPECTRUM'].data['WAVE']*10,
                                    mask=numpy.logical_not(numpy.isfinite(
                                                hdu['SPECTRUM'].data['WAVE'])))
        flux = numpy.ma.MaskedArray(hdu['SPECTRUM'].data['FLUX']*1e17,
                                    mask=numpy.logical_not(numpy.isfinite(
                                                hdu['SPECTRUM'].data['FLUX'])))
        ferr = numpy.ma.MaskedArray(hdu['SPECTRUM'].data['ERR']*1e17,
                                    mask=numpy.logical_not(numpy.isfinite(
                                                hdu['SPECTRUM'].data['ERR'])))
        if not minimum_useful:
            return wave, flux, ferr
        indx = wave > float(hdu[0].header['WMIN_UVB'])*10
        return wave[indx], flux[indx], ferr[indx]



def read_miles(ifile):
    with fits.open(ifile) as hdu:
        flux = hdu[0].data
        wave = numpy.arange(flux.size)*hdu[0].header['CDELT1'] + hdu[0].header['CRVAL1']
    return wave, flux


def test_rukia_model():

    sigma = LegendrePolynomial(0)
    mult = LegendrePolynomial(0)
    r = Rukia(sigma, mul_model=mult)

    # Build a fake spectrum
    dw = 1.0
    tpl_wave = numpy.arange(4500.0, 4600.0, dw)
    tpl_flux = numpy.zeros(tpl_wave.shape, dtype=float)
    sigma = 2.5/DAPConstants.sig2fwhm
    for c in numpy.linspace(tpl_wave[0], tpl_wave[-1], num=5)[1:-1]:
        tpl_flux += pixelated_gaussian(tpl_wave, c=c, s=sigma)/dw

    wave = numpy.arange(4500.0, 4600.0, dw) + 3.
    par = numpy.ones(r.np, dtype=float)
    par[0] = 0.
    model = r.model(par, wave=wave, tpl_wave=tpl_wave, tpl_flux=tpl_flux)

    sigma = numpy.sqrt(numpy.square(sigma) + 1)
    flux = numpy.zeros(wave.shape, dtype=float)
    for c in numpy.linspace(tpl_wave[0], tpl_wave[-1], num=5)[1:-1]:
        flux += pixelated_gaussian(wave, c=c, s=sigma)/dw

    assert numpy.amax(numpy.absolute(model-flux)) < 1e-4, 'Bad model construction'


def test_rukia_fit():
    miles_wave, miles_flux = read_miles(data_test_file('MILES_res2.50_star_m0505V.fits'))
    miles_flux /= numpy.median(miles_flux)

    xsl_wave, xsl_flux, _ = read_xsl(data_test_file('xsl_spectrum_X0360_uvb.fits'))
    xsl_flux /= numpy.ma.median(xsl_flux)

    # Setup the fitting functions
    sigma = LegendrePolynomial(1)
    mulp = LegendrePolynomial(5)
    r = Rukia(sigma, mul_model=mulp)

    # Perform the fit
    r.fit(miles_wave, miles_flux, xsl_wave.data, xsl_flux.data, shift=0.0, fit_shift=True,
          rejiter=-1)

    assert numpy.sum(r.gpm) == 2175, 'Change in the number of pixels rejected.'
    assert numpy.sqrt(numpy.mean(numpy.square((r.flux-r.model(r.par))[r.gpm]))) < 0.01, \
            'Fit quality changed.'





#!/usr/bin/env python3

"""
Dynamically build the rst documentation of the bitmasks.
"""

import os
import time

#-----------------------------------------------------------------------------

def write_parset(parset_class, opath, class_link=True):
    ofile = os.path.join(opath, '{0}.rst'.format(parset_class.__name__.lower()))
    lines = parset_class().to_rst_table(header=False, class_link=class_link, nested=False)
    with open(ofile, 'w') as f:
        f.write('\n'.join(lines))

if __name__ == '__main__':
    t = time.perf_counter()

    from pkg_resources import resource_filename
    root = os.path.dirname(resource_filename('mangadap', ''))
    path = os.path.join(root, 'docs', 'tables')
    if not os.path.isdir(path):
        os.makedirs(path)

    # Tables to write:
    #
#    from mangadap.config.analysisplan import AnalysisPlan
#    write_parset(AnalysisPlan, path, class_link=False)

    from mangadap.par.artifactdb import ArtifactPar
    write_parset(ArtifactPar, path, class_link=False)
    from mangadap.par.emissionlinedb import EmissionLinePar
    write_parset(EmissionLinePar, path, class_link=False)

    from mangadap.proc.reductionassessments import ReductionAssessmentDef
    write_parset(ReductionAssessmentDef, path, class_link=False)

    from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectraDef
    write_parset(SpatiallyBinnedSpectraDef, path, class_link=False)
    from mangadap.proc.spatialbinning import RadialBinningPar, VoronoiBinningPar, SquareBinningPar
    write_parset(RadialBinningPar, path, class_link=False)
    write_parset(VoronoiBinningPar, path, class_link=False)
    write_parset(SquareBinningPar, path, class_link=False)
    from mangadap.proc.spectralstack import SpectralStackPar
    write_parset(SpectralStackPar, path, class_link=False)

    from mangadap.proc.templatelibrary import TemplateLibraryDef
    write_parset(TemplateLibraryDef, path, class_link=False)
    from mangadap.proc.stellarcontinuummodel import StellarContinuumModelDef
    write_parset(StellarContinuumModelDef, path, class_link=False)
    from mangadap.proc.ppxffit import PPXFFitPar
    write_parset(PPXFFitPar, path, class_link=False)

    from mangadap.proc.emissionlinemoments import EmissionLineMomentsDef
    write_parset(EmissionLineMomentsDef, path, class_link=False)
    from mangadap.proc.bandpassfilter import BandPassFilterPar
    write_parset(BandPassFilterPar, path, class_link=False)

    from mangadap.proc.emissionlinemodel import EmissionLineModelDef
    write_parset(EmissionLineModelDef, path, class_link=False)
#    from mangadap.proc.elric import ElricPar
#    write_parset(ElricPar, path, class_link=False)
    from mangadap.proc.sasuke import SasukePar
    write_parset(SasukePar, path, class_link=False)

    from mangadap.proc.spectralindices import SpectralIndicesDef
    write_parset(SpectralIndicesDef, path, class_link=False)

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






#!/usr/bin/env python3

import os
import numpy

import astropy.constants
from astropy.io import fits

from mangadap.config import defaults
from mangadap.config import manga
from mangadap.datacube import MaNGADataCube
from mangadap.survey.manga_dap import manga_dap
from mangadap.par.analysisplan import AnalysisPlan, AnalysisPlanSet

#-----------------------------------------------------------------------------
def get_config(plt, ifu, config_file, drpall_file=None):
    if drpall_file is None:
        drpall_file = manga.drpall_file()

    # Use the DRPall file
    with fits.open(drpall_file) as hdu:
        indx = numpy.where(hdu['MANGA'].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu))[0]
        if len(indx) != 1:
            raise ValueError('{0}-{1} either does not exist or has more than one match!'.format(
                             plt, ifu))

        MaNGADataCube.write_config(config_file, plt, ifu, z=hdu[1].data['z'][indx[0]],
                                   ell=1-hdu[1].data['nsa_elpetro_ba'][indx[0]],
                                   pa=hdu[1].data['nsa_elpetro_phi'][indx[0]],
                                   reff=hdu[1].data['nsa_elpetro_th50_r'][indx[0]],
                                   overwrite=True)


def fit_one_cube(plt, ifu, drpall_file=None, directory_path=None, analysis_path=None):
    # Grab the required input parameters
    config_file = '{0}-{1}.cfg'.format(plt, ifu)
    get_config(plt, ifu, config_file, drpall_file=drpall_file)

    # Read the datacube
    cube = MaNGADataCube.from_config(config_file, directory_path=directory_path)

    # Define how you want to analyze the data
    plan = AnalysisPlanSet([ AnalysisPlan(drpqa_key='SNRG',
                                          bin_key='VOR10', #'HYB10',
                                          continuum_key='MILESHCMPL10',
                                          elmom_key='EMOMMPL10',
                                          elfit_key='EFITMPL10', #'EFITMPL9DB',
                                          spindex_key='INDXEN') ])

    # Run it!
    return manga_dap(cube, plan, verbose=2, directory_path=directory_path,
                     analysis_path=analysis_path)
#-----------------------------------------------------------------------------


if __name__ == '__main__':
    drpver = 'v3_0_1'
    directory_path = os.path.join(defaults.dap_source_dir(), 'data', 'remote')
    drpall_file = os.path.join(directory_path, 'drpall-{0}.fits'.format(drpver))
    fit_one_cube(7815, 3702, drpall_file=drpall_file, directory_path=directory_path,
                 analysis_path='./output')




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a class to handle an SDSS trace set; cf. pydl TraceSet.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import numpy

class TraceSet:
    """
    Interact with a trace set written by IDLUTILS.

    Args:
        func (:obj:`str`):
        xmin (scalar-like):
        xmax (scalar-like):
        coeff (array-like):

    """
    def __init__(self, func, xmin, xmax, coeff):
        if func not in self.allowed_functions():
            raise ValueError('{0} is not a valid function for use with TraceSet.'.format(func))
        self.func = func
        self.val, self.vander = self._get_val_and_vander()

        self.coeff = numpy.atleast_2d(coeff)
        if self.coeff.ndim != 2:
            raise ValueError('Coefficients array can have no more than two dimensions.')

        # The number of traces and the order of each trace
        self.size, self.order = self.coeff.shape
        self.order -= 1

        self.xmin = xmin
        self.xmax = xmax
        self.Dx = xmax-xmin
        self.xmid = xmin + self.Dx/2

    @classmethod
    def from_fits_table(cls, data):
        return cls(data['FUNC'][0], data['XMIN'][0], data['XMAX'][0], data['COEFF'][0])

    @staticmethod
    def allowed_functions():
        return ['poly', 'legendre', 'chebyshev']

    def _get_val_and_vander(self):
        if self.func == 'poly':
            return numpy.polynomial.polynomial.polyval, numpy.polynomial.polynomial.polyvander
        if self.func == 'legendre':
            return numpy.polynomial.legendre.legval, numpy.polynomial.legendre.legvander
        if self.func == 'chebyshev':
            return numpy.polynomial.chebyshev.chebval, numpy.polynomial.chebyshev.chebvander

        raise ValueError('{0} is not a valid function for use with TraceSet.'.format(self.func))

    def _normalized_coordinates(self, x):
        return 2 * (x-self.xmid)/self.Dx

    def sample(self, x, index=None):
        """
        Sample the trace at the provided coordinates.

        Args:
            x (array-like):
                One-dimensional array with coordinates to sample.

            index (:obj:`int`, optional):
                Sample a single trace with the provided index.  Default
                is to return an array sampling all traces.

        Returns:
            numpy.ndarray: A 1 or 2D array with the samples of one or
            more traces.
        """
        if index is not None:
            return self.val(self._normalized_coordinates(x), self.coeff[index,:])

        return numpy.dot(coeff, self.vander(self._normalized_coordinates(x), self.order))




#!/usr/bin/env python3
""" LambdaR for MaNGA
 Version 0.0.2  - Eric, after editing and adding the relevant functions
 Version 0.0.1  - Kyle 
"""

import sys

from os import remove
import os.path

import numpy
from time import clock
from mangadap.util.exception_tools import print_frame
from mangadap.dapfile import dapfile

from mangadap.contrib.LambdaR_2D_forMaNGA import Derive_LR_VS_Profiles, derive_unbinned_field

from matplotlib import pyplot

# Pull the necessary data out of the DAP file
def dap_data(dapf):
    ell = dapf.nsa_ellipticity()
    pa = dapf.guess_position_angle()
    xpos = dapf.hdu['DRPS'].data['XPOS']
    ypos = dapf.hdu['DRPS'].data['YPOS']
    signal = dapf.hdu['DRPS'].data['SIGNAL']
    binid = dapf.hdu['DRPS'].data['BINID']
    binx = dapf.hdu['BINS'].data['BINXRL']
    biny = dapf.hdu['BINS'].data['BINYRU']
    binf = dapf.hdu['BINS'].data['BINF']
    stellar_v = dapf.hdu['STFIT'].data['KIN'][:,0]
    stellar_sigma = dapf.hdu['STFIT'].data['KIN'][:,1]
    return ell, pa, xpos, ypos, signal, binid, binx, biny, stellar_v, stellar_sigma

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    # Start the clock
    t = clock()

    # Check number of command-line arguments is correct
    narg = 6
    if len(sys.argv) != narg:
        print('Usage: Test_LR_MaNGA <plate> <ifudesign> <mode (CUBE or RSS)> <bintype (e.g.' \
              ' STON)> <iter>')
        raise Exception('Incorrect number of arguments!')

    # Try to open the DAP output file
    try:
        dapf = dapfile(plate=sys.argv[1], ifudesign=sys.argv[2], mode=sys.argv[3],
                       bintype=sys.argv[4], niter=sys.argv[5], read=True)
    except Exception as e:
        print(e)
        print_frame('Exception')
        exit(1)

    # Try to open the associated parameter file
    try: 
        dapf.read_par()
    except Exception as e:
        print(e)
        print_frame('Exception')
        exit(1)

    # Get the necessary data
    ell, pa, xpos, ypos, signal, binid, binx, biny, stellar_v, stellar_sigma = dap_data(dapf)

    # Run Eric's functions to produce output data
    sel_spaxel = (binid >= 0)
    xunb, yunb, V_unb = derive_unbinned_field(binx, biny, stellar_v, 
                                              xpos[sel_spaxel], ypos[sel_spaxel])
    xunb, yunb, S_unb = derive_unbinned_field(binx, biny, stellar_sigma, 
                                              xpos[sel_spaxel], ypos[sel_spaxel])

    # Deriving the maximum Radius and creating an array to sample the radii for LambdaR
    Max_Radius = numpy.max(numpy.sqrt(xpos**2 + ypos**2))
    Min_Radius = 1.0  # Minimum Radius to consider [arcsec]
    N_Radius = 30 # number of points to consider 
    print('Maximum radius is {0}'.format(Max_Radius))

    Result = Derive_LR_VS_Profiles(xunb, yunb, signal[sel_spaxel], V_unb, S_unb, 
                                   Eps=ell, PA=pa, verbose=False, Min_Radius=Min_Radius,
                                   Max_Radius=Max_Radius, N_Radius=N_Radius)

    print('Maximum radius is {0}'.format(Result.Max_Radius))

    print(Result.R_EpsPA.shape)
    print(Result.SemiMajorRadius.shape)
    print(Result.LambdaR.shape)
    print(Result.VS.shape)

    pyplot.plot(Result.SemiMajorRadius, Result.LambdaR)
    pyplot.plot(Result.SemiMajorRadius, Result.VS)
    pyplot.show()

    # Write the output file(s)

    # Finish
    print('Elapsed time: {0} seconds'.format(clock() - t))

# dap/trunk/python/mangadap/contrib/dapcontrib_database.py 8143 12701 'CUBE' 'STON' 1
# dapf = dapfile(plate=8143, ifudesign=12701, mode='CUBE', bintype='STON', niter=1, analysis_path='/home/misc/ManGA/Data/manga/spectro/analysis/MPL-3/')



import pytest

from astropy.io import fits

from mangadap.par.parset import ParSet
from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionlinedb import EmissionLineDB
from mangadap.par.bandheadindexdb import BandheadIndexDB
from mangadap.par.absorptionindexdb import AbsorptionIndexDB
from mangadap.par.emissionmomentsdb import EmissionMomentsDB

#-----------------------------------------------------------------------------

def test_parset():

    # Instantiate a ParSet
    keys = [ 'test', 'par', 'list', 'junk']
    defaults = [ 'this', 0, 0.0, None ]
    options = [ 'this', None, None, None ]
    dtypes = [ str, int, [int, float], None ]
    descr = [ 'test parameter description',
              'par parameter description',
              None,
              'this is a rather long description of the junk parameter.  You can include ' \
                'rst-style references like pointing back to the ' \
                ':class:`mangadap.par.parset.ParSet` class, for when this description is ' \
                'written to an rst table using :func:`mangadap.par.parset.ParSet.to_rst_table` ' \
                'and included in an rst doc synthesized into html using sphinx.']
    par = ParSet(keys, defaults=defaults, options=options, dtypes=dtypes, descr=descr)

    # Print it
    par.info()
    print(par)

    # Test values
    assert all([ k in par.keys() for k in keys]), 'Not all keys assigned'
    assert par['junk'] is None, 'Default should be None'
    assert par['test'] == 'this', 'Bad instantiation'

    # Test constraints
    with pytest.raises(KeyError):
        par['that'] = 'new', 'This key does not exist; need to use add method.'
    with pytest.raises(ValueError):
        par['test'] = 'that', 'Options not correctly checked.'
    with pytest.raises(TypeError):
        par['par'] = 'test', 'Type not correctly checked'

    # Test IO
    ParSet.from_dict(par.to_dict())

    hdr = fits.Header()
    par.to_header(hdr)
    ParSet.from_header(hdr)

    ParSet.from_config(par.to_config())

    # Try adding a new parameter
    par.add('echo', 10, dtype=int)
    assert 'echo' in par.keys(), 'Key not added'
    with pytest.raises(TypeError):
        par['echo'] = 1.3, 'Type not added'


def test_artifact():

    artdb = ArtifactDB.from_key('BADSKY')
    assert artdb.size == 1

    _artdb = ArtifactDB(artdb.file)
    assert _artdb.size == 1


def test_emldb():
    emldb = EmissionLineDB.from_key('ELPMPL11')
    assert emldb.size == 35

    _emldb = EmissionLineDB(emldb.file)
    assert _emldb.size == 35


def test_eml_channel_names():
    emldb = EmissionLineDB.from_key('ELPMPL11')
    names = emldb.channel_names()
    assert 'Ha-6564' in list(names.keys())
    assert names['OII-3727'] == 0


def test_elmom():
    elmom = EmissionMomentsDB.from_key('ELBMPL9')
    assert elmom.size == 35

    _elmom = EmissionMomentsDB(elmom.file)
    assert _elmom.size == 35


def test_elmom_channel_names():
    elmom = EmissionMomentsDB.from_key('ELBMPL9')
    names = elmom.channel_names()
    assert 'Ha-6564' in list(names.keys())
    assert names['OIId-3728'] == 0


def test_bhddb():

    bhddb = BandheadIndexDB.from_key('BHBASIC')
    assert bhddb.size == 3

    _bhddb = BandheadIndexDB(bhddb.file)
    assert _bhddb.size == 3


def test_bhd_channel_names():
    bhddb = BandheadIndexDB.from_key('BHBASIC')
    names = bhddb.channel_names()
    assert 'D4000' in list(names.keys())
    assert names['D4000'] == 0


def test_absdb():

    absdb = AbsorptionIndexDB.from_key('LICKINDX')
    assert absdb.size == 21

    _absdb = AbsorptionIndexDB(absdb.file)
    assert _absdb.size == 21


def test_abs_channel_names():
    absdb = AbsorptionIndexDB.from_key('LICKINDX')
    names = absdb.channel_names()
    assert 'CN1' in list(names.keys())
    assert names['CN1'] == 0









def get_manga_environ():

    import os
    from ..version import version

    # Default versions
    _MANGADRP_VER = 'v3_1_1'
    _MANGACORE_VER = 'v1_9_1'

    # First set the default environment
    environ = {'MANGADRP_VER': _MANGADRP_VER,
               'MANGA_SPECTRO_REDUX': os.path.join(os.getcwd(), 'MaNGA', 'redux'),
               'MANGADAP_VER': version,
               'MANGA_SPECTRO_ANALYSIS': os.path.join(os.getcwd(), 'MaNGA', 'analysis'),
               'MANGACORE_VER': _MANGACORE_VER,
               'MANGACORE_DIR': os.path.join(os.getcwd(), 'MaNGA', 'core', _MANGACORE_VER)}

    # Replace with variables from working environment if they exist
    for key in environ.keys():
        if key in os.environ:
            environ[key] = os.environ[key]

    environ['sdss-access'] = os.environ['SDSS_ACCESS_DIR'].split('/')[-1] \
                                if 'SDSS_ACCESS_DIR' in os.environ else None
    environ['idlutils'] = os.environ['IDLUTILS_DIR'].split('/')[-1] \
                            if 'IDLUTILS_DIR' in os.environ else None

    return environ

manga_environ = get_manga_environ()

def get_python_versions():

    versions = {}
    import sys
    versions['python'] = '.'.join([str(v) for v in sys.version_info[:3]])
    import numpy
    versions['numpy'] = numpy.__version__
    import scipy
    versions['scipy'] = scipy.__version__
    import matplotlib
    versions['matplotlib'] = matplotlib.__version__
    import astropy
    versions['astropy'] = astropy.__version__
    import pydl
    versions['pydl'] = pydl.__version__
    import ppxf
    versions['ppxf'] = ppxf.__version__
    import vorbin
    versions['vorbin'] = vorbin.__version__
    from ..version import version
    versions['mangadap'] = version
    return versions

python_versions = get_python_versions()




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Defines a class used to produce and provide an interface for data
required to create :ref:`execution-config`.

The drpcomplete file/data is primarily required for the survey-level
:ref:`execution` of the DAP. However, it does provide useful information
regarding the completed DRP files.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
from pathlib import Path
import os
import warnings

from IPython import embed

import numpy

from astropy.io import fits
import astropy.constants

from pydl.pydlutils.yanny import yanny

from ..datacube import MaNGADataCube
from ..spectra import MaNGARSS
from ..config import manga
from ..util.parser import arginp_to_list

class DRPComplete:
    r"""
    Database with information needed by the DAP to analyze the completed DRP
    files.

    This class searches the defined paths for files resulting from the 3D phase
    of the MaNGA DRP, and then collates the information necessary to run those
    files through the MaNGA DAP.  The necessary metadata are pulled from the
    DRPall file; see :func:`update`.

    Args:
        platelist (:obj:`str`, :obj:`list`, optional):
            List of plates to search for.  Default is to search the full DRP
            path.
        ifudesignlist (:obj:`str`, :obj:`list`, optional):
            List of IFUs to search for.  Default is to search the full DRP path.
        drpall (:obj:`str`, optional):
            The full path to the DRPall fits file.  Default is set by
            :func:`mangadap.config.manga.drpall_file`.
        drpver (:obj:`str`, optional):
            DRP version, which is:

                - used to define the default DRP redux path
                - used in the name of the drpcomplete fits file
                - included as a header keyword in the output file

            Default is defined by :func:`mangadap.config.manga.drp_version`.
        redux_path (:obj:`str`, optional): 
            The path to the top-level directory containing the DRP output files.
            Default is defined by :func:`mangadap.config.manga.drp_redux_path`.
        dapver (:obj:`str`, optional):
            DAP version, which is:

                - used to define the default DAP analysis path
                - included as a header keyword in the output drpcomplete
                  fits file

            Default is defined by :func:`mangadap.config.manga.dap_version`.
        analysis_path (:obj:`str`, optional):
            The path to the top level directory for the DAP output files.
            Default is defined by
            :func:`mangadap.config.manga.dap_analysis_path`
        directory_path (:obj:`str`, optional):
            Direct path for the output file.  Default is a directory called
            ``common`` within the ``analysis_path`` directory.
        readonly (:obj:`bool`, optional):
            Flag that the drpcomplete fits file is only opened for
            reading, not for updating.  If True, many of the attributes
            will be set to None.

    Attributes:
        platelist (:obj:`list`):
            List of plates to search for, see above.
        ifudesignlist (:obj:`list`):
            List of IFU designs to search for, see above.
        drpall (:obj:`str`):
            The DRPall file, see above.
        drpver (:obj:`str`):
            DRP version, see above.
        redux_path (`Path`_):
            Path to the top level of directory for the DRP output, see
            above.
        dapver (:obj:`str`):
            DAP version, see above.
        analysis_path (`Path`_):
            Path to the top level directory for the DAP output files,
            see above.
        directory_path (`Path`_):
            Direct path for the output file.
        readonly (:obj:`bool`):
            Flag that the drpcomplete fits file is only opened for
            reading, not for updating.
        hdu (`astropy.io.fits.HDUList`_):
            Fits data with binary table data.
        nobs (:obj:`int`):
            Number of observations in the file

    """
    def __init__(self, platelist=None, ifudesignlist=None, drpall=None, drpver=None,
                 redux_path=None, dapver=None, analysis_path=None, directory_path=None,
                 readonly=False):

        # Input properties
        self.drpver = manga.drp_version() if drpver is None else str(drpver)
        self.redux_path = manga.drp_redux_path(self.drpver) \
                                if redux_path is None else Path(redux_path).resolve()

        self.dapver = manga.dap_version() if dapver is None else str(dapver)
        self.analysis_path = manga.dap_analysis_path(self.drpver, self.dapver) \
                             if analysis_path is None else Path(analysis_path).resolve()
        self.directory_path = self.analysis_path / 'common' if directory_path is None \
                                    else Path(directory_path).resolve()

        self.hdu = None
        self.nobs = None
        if self.file_path.exists():
            print('READING')
            self._read()

        if readonly:
            self.readonly=True
            self.platelist = None
            self.ifudesignlist = None
            return

        self.readonly = False

        self.platelist = arginp_to_list(platelist, evaluate=True)
        self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)

        self.drpall = manga.drpall_file(drpver=self.drpver, redux_path=self.redux_path) \
                            if drpall is None else drpall
        
    def __getitem__(self, key):
        return self.hdu['DRPC'].data[key]

    def __len__(self):
        return self.nobs

    # ******************************************************************
    #  Utility functions
    # ******************************************************************
    def _read_fix_data(self, fix_file):
        """
        Read "fix" data from the provided file.

        Fix data are generally corrections to the metadata provides by the NSA
        for specific MaNGA observations.

        Args:
            fix_file (:obj:`str`):
                SDSS parameter file with the fix data.

        Returns:
            yanny: A yanny structure with the data fixes.  Returns
            None and raises a warning if the file does not exist.
        """
        if not fix_file.exists():
            warnings.warn(f'DAP "fix" file does not exist: {fix_file}')
            return None
        return yanny(filename=str(fix_file))

    def _read_redshift_fix(self):
        """
        Wrapper for :func:`_read_fix_data` and
        :func:`~mangadap.config.defaults.redshift_fix_file` that
        returns the redshift fix data.
        """
        return self._read_fix_data(manga.redshift_fix_file())

    def _read_photometry_fix(self):
        """
        Wrapper for :func:`_read_fix_data` and
        :func:`~mangadap.config.defaults.photometry_fix_file` that
        returns the photometry fix data.
        """
        return self._read_fix_data(manga.photometry_fix_file())

    def _match_drpall(self):
        """
        Find the data needed for the DRP complete database in the DRPall
        file.

        If a plate-ifudesign combination is not found in the DRPall
        file, the other parameter values are set to -9999 and the MaNGA
        ID is set to NULL.

        If the plate-ifudesign combination is found, the columns in the
        plateTargets files that the code expects to find are ``'plate'``,
        ``'ifudsgn'``, ``'mangaid'``, ``'objra'``, ``'objdec'``; the values that
        will be replaced with ``'NULL'`` (str) or -9999 (int or float) if they
        do not exist are ``'nsa_version'``, ``'nsa_nsaid'``, ``'mngtarg1'``,
        ``'mngtarg3'``, ``'z'``, ``'nsa_elpetro_ba'``, ``'nsa_elpetro_phi'``,
        ``'nsa_elpetro_th50_r'``, ``'nsa_vdisp'``.
        
        Returns:
            :obj:`tuple`: 14 `numpy.ndarray`_ arrays with the MaNGA ID, object
            right ascension, object declination, catalog ID, index of the entry
            in the catalog, catalog version, ID of object in the catalog, main
            MaNGA survey target bitmask, ancillary MaNGA survey target bitmask,
            velocity, velocity dispersion, ellipticity, position angle, and
            effective radius.
        """
        if not self.drpall.exists():
            raise ValueError(f'DRPall file does not exist! {self.drpall}')
        # Open the drpall file
        hdu = fits.open(str(self.drpall))
        # Check the plateifus are unique
        pltifu = hdu[1].data['PLATEIFU']
        if len(numpy.unique(pltifu)) != len(pltifu):
            raise ValueError('The PLATEIFU column in the DRPall file is not unique!')

        # Find the rows in the DRPall file with the correct; assumes the 
        this_pltifu = numpy.array([f'{p}-{i}' for p,i in zip(self.platelist,self.ifudesignlist)])
        rows = numpy.array([numpy.where(pltifu == pi)[0][0] if pi in pltifu else -1 \
                                for pi in this_pltifu])
        if numpy.any(rows < 0):
            raise ValueError('Could not find the following in the DRPall file: '
                             '{0}'.format(this_pltifu[rows < 0]))
        indx = rows > -1
        rows = rows[indx]

        # Initialize the data
        n_drp = len(self.platelist)
        mangaid = numpy.full(n_drp, 'NULL', dtype=object)
        objra = numpy.full(n_drp, -9999.0, dtype=float)
        objdec = numpy.full(n_drp, -9999.0, dtype=float)
        catid = numpy.full(n_drp, -9999, dtype=int)
        catindx = numpy.full(n_drp, -9999, dtype=int)
        trg_version = numpy.full(n_drp, 'NULL', dtype=object)
        trg_id = numpy.full(n_drp, -9999, dtype=int)
        manga_trg1 = numpy.full(n_drp, -9999, dtype=int)
        manga_trg3 = numpy.full(n_drp, -9999, dtype=int)
        vel = numpy.full(n_drp, -9999.0, dtype=float)
        veldisp = numpy.full(n_drp, -9999.0, dtype=float)
        ell = numpy.full(n_drp, -9999.0, dtype=float)
        pa = numpy.full(n_drp, -9999.0, dtype=float)
        Reff = numpy.full(n_drp, -9999.0, dtype=float)

        # Get the data
        mangaid[indx] = hdu[1].data['mangaid'][rows]
        objra[indx] = hdu[1].data['objra'][rows]
        objdec[indx] = hdu[1].data['objdec'][rows]
        for i in numpy.where(indx)[0]:
            catid[i], catindx[i] = map(lambda x: x.astype(int).tolist(),
                                        numpy.array(mangaid[i].split('-')))
        trg_version[indx] = hdu[1].data['nsa_version'][rows]
        trg_id[indx] = hdu[1].data['nsa_nsaid'][rows]
        manga_trg1[indx] = hdu[1].data['mngtarg1'][rows]
        manga_trg3[indx] = hdu[1].data['mngtarg3'][rows]
        vel[indx] = hdu[1].data['z'][rows] * astropy.constants.c.to('km/s').value
        # There is no vdisp
        ell[indx] = 1 - hdu[1].data['nsa_elpetro_ba'][rows]
        pa[indx] = hdu[1].data['nsa_elpetro_phi'][rows]
        Reff[indx] = hdu[1].data['nsa_elpetro_th50_r'][rows]

        # Done
        hdu.close()
        return mangaid.astype(str), objra, objdec, catid, catindx, trg_version.astype(str), \
                    trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa, Reff

    def _all_data_exists(self):
        """
        Determines if the data for all the plates/ifudesigns selected in
        the current compilation is already present in the current
        drpcomplete fits file.

        Returns:
            :obj:`bool`: Flag that all data has already been collated
            for the requested plate/ifudesign list.
        """
        if not self._confirm_access():
            return False
        for p,i in zip(self.platelist, self.ifudesignlist):
            if numpy.sum((self['PLATE'] == p) & (self['IFUDESIGN'] == i)) == 0:
                return False
        return True

    def _read(self):
        """Read the data in the existing file at :func:`file_path`."""
        if self.hdu is not None:
            self.hdu.close()
            self.hdu = None
        self.hdu = fits.open(str(self.file_path))
        self.nobs = self.hdu['DRPC'].header['NAXIS2']
        print('Read data: {0} rows'.format(self.nobs))

    def _confirm_access(self, reread=False):
        """
        Check the drpcomplete fits file at :func:`file_path` is
        accessible, and read the data if not yet read.

        Args:
            reread (:obj:`bool`, optional):
                Force the file to be re-read
        
        Returns:
            :obj:`bool`: Flag that :attr:`hdu` is read and valid.
        """
        if not self.file_path.exists():
            return False
        if self.hdu is None or reread:
            self._read()
        return True

    def _find_completed_reductions(self, mindesign=19, combinatorics=False, on_disk=False):
        """
        Search the DRP path for reduced CUBE files.

        Function allows for one or both of :attr:`platelist` and
        :attr:`ifudesignlist` to be None.  The behavior is:
        
            - If both are None, all available CUBE files are used to
              create :attr:`platelist` and :attr:`ifudesignlist`.

            - If one is None, all available CUBE files within the
              provided list of the one that is not None are used to
              create :attr:`platelist` and :attr:`ifudesignlist`.  E.g.,
              if :attr:`platelist` =[7443] and :attr:`ifudesignlist` is
              None, all CUBE files with plate=7443 are chosen.

            - If both are not None and they have different lengths, or
              the same length and combinatorics is True, all available
              CUBE files within the constraints of both lists are
              selected. E.g., if :attr:`platelist` =[7443,7495] and
              :attr:`ifudesignlist` =[12704], the CUBE files with
              (plate,ifudesign)=[(7443,12704),(7459,12704)] are chosen.

            - If both are not None and they have the same length and
              combinatorics is False, all available CUBE files in the
              matched lists are chosen.  E.g. if :attr:`platelist`
              =[7443,7495] and :attr:`ifudesignlist` =[12704,12703], the
              CUBE files with
              (plate,ifudesign)=[(7443,12704),(7459,12703) are chosen

        Args:
            mindesign (:obj:`int`, optional):
                Minimum bundle design to consider.  For example, to
                ignore all the 7-fiber bundles, set `mindesign=19`
                to only select the bundles with 19 or more fibers.
            combinatorics (:obj:`bool`, optional):
                Create :attr:`platelist` and :attr:`ifudesignlist` by
                determining all possible combinations of the input
                values. See above.
            on_disk (:obj:`bool`, optional):
                When searching for available files to analyze, search
                the DRP directory path instead of using the data in the
                DRPall file.

        Returns:
            :obj:`tuple`: Two lists with the available plates and ifudesigns for
            which to collect data.
        """
        # Are the plate and IFU lists matched?
        matchedlist = (self.platelist is not None and self.ifudesignlist is not None \
                       and (len(self.platelist) == len(self.ifudesignlist) and not combinatorics))

        # Get the list of files
        if matchedlist:
            # Lists already matched, just construct the file names
            files = [manga.MaNGAConfig(int(p), int(i), drpver=self.drpver,
                                       redux_path=self.redux_path).file_path
                        for p,i in zip(self.platelist, self.ifudesignlist)]
        elif on_disk:
            # Find the DRP LOGCUBE files on disk
            print('Searching for completed DRP CUBE files...', end='\r')
            files = sorted(list(self.redux_path.glob('*/stack/*-LOGCUBE.fits.gz')))
            print('Searching for completed DRP CUBE files...DONE.')
        else:
            # Use the DRPall file
            drpall_hdu = fits.open(str(self.drpall))
            pltifu = drpall_hdu[1].data['PLATEIFU']
            if len(numpy.unique(pltifu)) != len(pltifu):
                raise ValueError('The PLATEIFU column in the DRPall file is not unique!')

            files = [manga.MaNGAConfig(int(p), int(i), drpver=self.drpver,
                                       redux_path=self.redux_path).file_path
                        for p,i in zip(drpall_hdu[1].data['plate'], drpall_hdu[1].data['ifudsgn'])]

        files = [f for f in files if f.exists()]
        if len(files) == 0:
            raise ValueError('Could not find any files to analyze.')

        # Get the list of plate and ifus
        pltifu = numpy.array([manga.parse_plate_ifu_from_file(f.name) for f in files]).astype(int)

        # Ignore any below the minimum ifusize
        pltifu = pltifu[numpy.logical_not(pltifu[:,1]//100 < mindesign),:]
        if matchedlist:
            return pltifu[:,0], pltifu[:,1]

        # Select those that match to the provided list
        indx = numpy.ones(len(pltifu), dtype=bool) if self.platelist is None \
                    else numpy.array([p in self.platelist for p in pltifu[:,0]])
        indx &= numpy.ones(len(pltifu), dtype=bool) if self.ifudesignlist is None \
                    else numpy.array([i in self.ifudesignlist for i in pltifu[:,1]])
        return pltifu[indx,0], pltifu[indx,1]
        
    def _find_modes(self):
        """
        Using the provided list of CUBE DRP files, find if the RSS mode
        is also available.

        .. warning::
            - This assumes everything in :attr:`platelist` and
              :attr:`ifudesignlist` has a 'CUBE' file.

        Currently only two modes are possible:

            - (1) only 'CUBE' file are available

            - (2) both 'CUBE' and 'RSS' files are available

        Returns:
            `numpy.ndarray`: Array of modes for each input DRP file.
        """
        print('Checking for RSS counterparts...', end='\r')
        has_rss = [manga.MaNGAConfig(int(p), int(i), mode='RSS', drpver=self.drpver,
                                       redux_path=self.redux_path).file_path.exists()
                        for p,i in zip(self.platelist, self.ifudesignlist)]
        print('Checking for RSS counterparts...DONE.')
        modes = numpy.ones(len(has_rss), dtype=int)
        modes[has_rss] = 2
        return modes
        
    # ******************************************************************
    #  User functions
    # ******************************************************************
    @property
    def file_name(self):
        """Return the name of the DRP complete database."""
        return f'drpcomplete_{self.drpver}.fits'

    @property
    def file_path(self):
        """Return the full pat to the DRP complete database."""
        return self.directory_path / self.file_name

    def update(self, platelist=None, ifudesignlist=None, combinatorics=False, force=False,
               alldrp=False, on_disk=False, quiet=False):
        """
        Update the DRP complete file.
        
        If `platelist` and/or `ifudesignlist` are provided, the existing
        :attr:`platelist` and :attr:`ifudesignlist` attributes are
        replaced.

        If `platelist` and `ifudesignlist` do not have the same length
        or `combinatorics` is True, the lists are expanded to include
        all combinations of their elements.

        If `platelist` and `ifudesignlist` are None or `alldrp` is True,
        all available plates/ifudesigns are collected from within the
        DRP directory structure.

        If the result of :func:`file_path` does not exist, it is
        created.

        If the result of :func:`file_path` does exist, the available
        plates and ifudesigns in the file are compared against the list
        (provided or collected) to update.  If the lists differ, the
        drpcomplete fits file is re-created from scratch.  If all the
        plates and ifudesigns are available, nothing is done, unless
        `force=True`.

        Args:
            platelist (:obj:`str`, :obj:`list`, optional):
                List of plates to include in the drpcomplete fits file.
            ifudesignlist (:obj:`str`, :obj:`list`, optional):
                List of ifudesigns to include in the drpcomplete fits
                file.
            combinatorics (:obj:`bool`, optional):
                Determine all combinations of the entered plates and
                ifudesigns.
            force (:obj:`bool`, optional):
                Overwrite any existing drpcomplete fits file with a new
                one built from scratch.
            alldrp (:obj:`bool`, optional):
                Find the full list of available DRP files.
            on_disk (:obj:`bool`, optional):
                When searching for available files to analyze, search
                the DRP directory path instead of using the data in the
                DRPall file.
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Raises:
            ValueError:
                Raised if drpcomplete fits file was opened in read-only
                mode.

        """
        if self.readonly:
            raise ValueError('drpcomplete fits file was opened as read-only!')

        if platelist is not None:
            self.platelist = arginp_to_list(platelist, evaluate=True)
        if ifudesignlist is not None:
            self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)

        if alldrp:
            self.platelist = None
            self.ifudesignlist = None

        # This *only* searches for CUBE files
        self.platelist, self.ifudesignlist \
                = self._find_completed_reductions(combinatorics=combinatorics, on_disk=on_disk)

        # Check if these plate-ifus are already in the table.
        if not force and self._all_data_exists():
            print(f'{self.file_path} up to date.')
            return
        else:
            print(f'Updating {self.file_path}.')

        # Past this point, the drpcomplete fits file will be ovewritten
        # if it already exists.  Only DRP files defined by 
        # self.platelist and self.ifudesignlist will be included; i.e.,
        # this does *not* append to the file but completely overwrites
        # it.

        # Check for the RSS files
        modes = self._find_modes()

        # Notify
        print('Number of DRP files for DRPComplete file: {0}'.format(len(self.platelist)))

        # Get the data
        mangaid, objra, objdec, catid, catindx, trg_version, trg_id, manga_trg1, manga_trg3, vel, \
                veldisp, ell, pa, Reff = self._match_drpall()

        # Apply any corrections to the redshifts using the redshift fix file
        redshift_fix_data = self._read_redshift_fix()
        if redshift_fix_data is not None:
            fix_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(redshift_fix_data['DAPZCORR']['plate'],
                                                   redshift_fix_data['DAPZCORR']['ifudesign'])])
            this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                       for p,i in zip(self.platelist,self.ifudesignlist)])
            rows = numpy.array([numpy.where(fix_pltifu == pi)[0][0] if pi in fix_pltifu else -1 \
                                for pi in this_pltifu])
            indx = rows > -1
            rows = rows[indx]
            vel[indx] = redshift_fix_data['DAPZCORR']['z'][rows] \
                            * astropy.constants.c.to('km/s').value

        photometry_fix_data = self._read_photometry_fix()
        if photometry_fix_data is not None:
            fix_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(photometry_fix_data['DAPPHOTCORR']['plate'],
                                                   photometry_fix_data['DAPPHOTCORR']['ifudesign'])])
            this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                       for p,i in zip(self.platelist,self.ifudesignlist)])
            rows = numpy.array([numpy.where(fix_pltifu == pi)[0][0] if pi in fix_pltifu else -1 \
                                for pi in this_pltifu])
            indx = rows > -1
            rows = rows[indx]
            ell[indx] = photometry_fix_data['DAPPHOTCORR']['ell'][rows]
            pa[indx] = photometry_fix_data['DAPPHOTCORR']['pa'][rows]
            Reff[indx] = photometry_fix_data['DAPPHOTCORR']['reff'][rows]

        # Write the data to disk
        self.write(self.platelist, self.ifudesignlist, modes, mangaid, objra, objdec, catid,
                   catindx, trg_version, trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa,
                   Reff)

    def write(self, platelist, ifudesignlist, modes, mangaid, objra, objdec, catid, catindx,
              trg_version, trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa, Reff, drpver=None,
              redux_path=None, dapver=None, analysis_path=None, overwrite=True):
        r"""
        Write the drpcomplete fits binary table.

        Header keywords are:

            - VERSDRP: DRP version
            - RDXPTH: DRP reduction path
            - VERSDAP: DAP version
            - SISPTH: DAP analysis path

        Binary table columns:

            1. PLATE: Plate number
            2. IFUDESIGN: IFU design
            3. MODES: Modes available 

                - MODES=1: only 'CUBE' file is available
                - MODES=2: both 'CUBE' and 'RSS' files are available

            4. MANGAID: MaNGA ID (same as 'CATID-CATINDX')
            5. OBJRA: Object right ascension
            6. OBJDEC: Object declination
            7. CATID: Catalog ID used for target selection
            8. CATINDX: Index in catalog with target data
            9. TRG_VERSION: Version of the catalog (e.g., NSA)
                used in targetting (def='NULL' if not available)
            10. TRG_ID: Target ID in catalog (def=-9999 if not available)
            11. MANGA_TARGET1: Main MaNGA survey target bitmask
            12. MANGA_TARGET3: Ancillary MaNGA survey target bitmask
            13. VEL: Systemic velocity from catalog (def=-9999.0 if
                not available)
            14. VDISP: Velocity dispersion from catalog
                (def=-9999.0 if not available)
            15. ELL: Ellipticity (1-b/a) from catalog (def=-9999.0
                if not available); Elliptical Petrosian value from NSA
            16. PA: Position angle from catalog (def=-9999.0 if not
                available); Elliptical Petrosian value from NSA
            17. REFF: Effective radius from catalog (def=-9999.0 if
                not available); Elliptical Petrosian value from NSA

        Args:
            platelist (:obj:`list`):
                List of plates
            ifudesignlist (:obj:`list`):
                List of IFU designs
            modes (`numpy.ndarray`_):
                Mode values, see above
            mangaid (`numpy.ndarray`_):
                MaNGA IDs
            objra (numpy.ndarray`_):
                Object right ascensions
            objdec (numpy.ndarray`_):
                Object declinations
            catid (numpy.ndarray`_):
                Catalog ID used for target selection
            catindx (numpy.ndarray`_):
                Index in catalog with target data
            trg_version (numpy.ndarray`_):
                Version of the catalog (e.g., NSA) used in targetting
            trg_id (numpy.ndarray`_):
                Target ID in catalog
            manga_trg1 (numpy.ndarray`_):
                Main MaNGA survey target bitmask
            manga_trg3 (numpy.ndarray`_):
                Ancillary MaNGA survey target bitmask
            vel (numpy.ndarray`_):
                Velocity from catalog, if available
            veldisp (numpy.ndarray`_):
                Velocity dispersion from catalog, if available
            ell (numpy.ndarray`_):
                Ellipticity (1-b/a) from catalog, if available
            pa (numpy.ndarray`_):
                Position angle from catalog, if available
            Reff (numpy.ndarray`_):
                Effective radius from catalog, if available
            drpver (:obj:`str`, optional):
                DRP version, see above.
            redux_path (`Path`_, optional):
                Path to the top level of directory for the DRP output, see
                above.
            dapver (:obj:`str`, optional):
                DAP version, see above.
            analysis_path (`Path`_, optional):
                Path to the top level directory for the DAP output files, see
                above.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing file.

        Raises:
            AttributeError:
                Raised if drpcomplete fits file was opened in read-only mode.
            FileExistsError:
                Raised if the drpcomplete file exists and overwrite=False.
        """
        if self.readonly:
            raise AttributeError('drpcomplete fits file was opened as read-only!')

        out = self.file_path
        if out.exists() and not overwrite:
            raise FileExistsError(f'DRP complete file already exists: {out}')

        if not self.directory_path.exists():
            self.directory_path.mkdir(parents=True)

        hdr = fits.Header()
        hdr['VERSDRP'] = self.drpver if drpver is None else drpver
        hdr['RDXPTH'] = str(self.redux_path) if redux_path is None else str(redux_path)
        hdr['VERSDAP'] = self.dapver if dapver is None else dapver
        hdr['SISPTH'] = str(self.analysis_path) if analysis_path is None else str(analysis_path)
        hdr['AUTHOR'] = 'K.B. Westfall <kbwestfall@gmail.com>'

        # TODO: Convert this into a DataTable?
        # Create the Binary Table
        cols = []
        cols.append(fits.Column(name='PLATE', format='1J', array=numpy.array(platelist)))
        cols.append(fits.Column(name='IFUDESIGN', format='1J', array=numpy.array(ifudesignlist)))
        cols.append(fits.Column(name='MODES', format='1B', array=modes))
        cols.append(fits.Column(name='MANGAID', format='{0}A'.format(mangaid.dtype.itemsize),
                                array=mangaid))
        cols.append(fits.Column(name='OBJRA', format='1D', array=objra))
        cols.append(fits.Column(name='OBJDEC', format='1D', array=objdec))
        cols.append(fits.Column(name='CATID', format='1J', array=catid))
        cols.append(fits.Column(name='CATINDX', format='1J', array=catindx))
        cols.append(fits.Column(name='TRG_VERSION',
                                format='{0}A'.format(trg_version.dtype.itemsize),
                                array=trg_version))
        cols.append(fits.Column(name='TRG_ID', format='1J', array=trg_id))
        cols.append(fits.Column(name='MANGA_TARGET1', format='1J', array=manga_trg1))
        cols.append(fits.Column(name='MANGA_TARGET3', format='1J', array=manga_trg3))
        cols.append(fits.Column(name='VEL', format='1D', array=vel))
        cols.append(fits.Column(name='VDISP', format='1D', array=veldisp))
        cols.append(fits.Column(name='ELL', format='1D', array=ell))
        cols.append(fits.Column(name='PA', format='1D', array=pa))
        cols.append(fits.Column(name='REFF', format='1D', array=Reff))

        # Create the HDUList and write it to the fits file
        self.hdu = fits.HDUList([fits.PrimaryHDU(header=hdr),
                                 fits.BinTableHDU.from_columns(cols, name='DRPC')])
        print(f'Writing to disk: {out}')
        self.hdu.writeto(str(out), overwrite=overwrite)
        self.nobs = self.hdu['DRPC'].header['NAXIS2']

    def write_config(self, ofile, plate=None, ifudesign=None, index=None, sres_ext=None,
                     sres_fill=None, covar_ext=None, reread=False, overwrite=True):
        """
        Write a config file with the data used to instantiate a
        :class:`mangadap.datacube.manga.MaNGADataCube` datacube for
        analysis.

        Args: 
            ofile (:obj:`str`, `Path`_):
                Output file name.
            plate (:obj:`int`, optional):
                Plate number.
            ifudesign (:obj:`int`, optional):
                IFU design.
            index (:obj:`int`, optional):
                Index of the row in :attr:`data` with the data to
                return.
            sres_ext (:obj:`str`, optional):
                The extension to use when constructing the spectral
                resolution vectors for the MaNGA datacubes. See
                :func:`mangadap.datacube.manga.MaNGADataCube.spectral_resolution`.
            sres_fill (:obj:`bool`, optional):
                Fill masked spectral-resolution data by simple linear
                interpolation.
            covar_ext (:obj:`str`, optional):
                Extension in the MaNGA DRP CUBE file to use as the
                single spatial correlation matrix for all wavelength
                channels.
            reread (:obj:`bool`, optional):
                Force the database to be re-read
            overwrite (:obj:`bool`, optional):
                Overwrite any existing parameter file

        Raises:
            IOError:
                Raised if the parameter file already exists and
                overwrite is False.
            ValueError:
                Raised if
                    - the row with the data is unknown because either
                      ``index`` is not defined or one or both of
                      ``plate`` and ``ifudesign`` is not defined.
                    - ``index`` does not exist in the data array.

        """
        if not self._confirm_access(reread=reread):
            raise IOError('Could not access database!')

        if (plate is None or ifudesign is None) and index is None:
            raise ValueError('Must provide plate and ifudesign or row index!')

        if index is None:
            index = self.entry_index(plate, ifudesign, reread=reread)
        elif index >= self.nobs:
            raise ValueError('Selected row index does not exist')

        manga.MaNGAConfig.write_config(ofile, self['PLATE'][index], self['IFUDESIGN'][index],
                                       log=True,
                                       z=self['VEL'][index]/astropy.constants.c.to('km/s').value,
                                       vdisp=self['VDISP'][index], ell=self['ELL'][index],
                                       pa=self['PA'][index], reff=self['REFF'][index],
                                       sres_ext=sres_ext, sres_fill=sres_fill, covar_ext=covar_ext,
                                       drpver=self.drpver, redux_path=self.redux_path,
                                       overwrite=overwrite)

    def entry_index(self, plate, ifudesign, reread=False):
        """
        Find the index of the row with the parameter data for the
        specified plate and ifudesign.

        .. warning::
            - This is very inefficient if you're looking for multiple
              entries...

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design
            reread (:obj:`bool`, optional):
                Force the database to be re-read

        Returns:
            :obj:`int`: Index of the row in :attr:`data` with the data for the
            given ``plate`` and ``ifudesign``.

        Raises:
            ValueError:
                Raised if the given ``plate`` and ``ifudesign`` were not found.
        """
        if not self._confirm_access(reread=reread):
            raise IOError('Could not access database!')
        indx = (self['PLATE'] == plate) & (self['IFUDESIGN'] == ifudesign)
        if numpy.sum(indx) == 0:
            raise ValueError(f'plate={plate},ifudesign={ifudesign} not in drpcomplete file!')
        return numpy.where(indx)[0][0]

    def can_analyze(self, row=None):
        """
        Determine if the DAP can analyze a plate-ifu entry in the
        database.

        The selection is currently:

            - MaNGAID != 'NULL'
            - MANGA_TARGET1 > 0 or MANGA_TARGET3 > 0
            - VEL > -500

        Args:
            row (:obj:`int`, optional):
                The specific row to test.  By default, return a boolean
                vector for all the database rows.

        Returns:
            Either a single boolean or boolean `numpy.ndarray`_
            flagging that DAP can (True) or cannot (False) analyze
            the data associated with the database entry (or entries).
        """
        if row is None:
            return (self['MANGAID'] != 'NULL') \
                    & ((self['MANGA_TARGET1'] > 0) | (self['MANGA_TARGET3'] > 0)) \
                    & (self['VEL'] > -500.0)
        return self['MANGAID'][row] != 'NULL' \
                and (self['MANGA_TARGET1'][row] > 0 or self['MANGA_TARGET3'][row] > 0) \
                and self['VEL'][row] > -500.0

